"""A package level configuration module."""
from pathlib import Path

NUM_SAMPLES = 1000
"""Number of cross section perturbation factors available in SCALE."""

SDF_DATA_NAMES = [
    "isotope",
    "reaction_type",
    "zaid",
    "reaction_mt",
    "zone_number",
    "zone_volume",
    "energy_integrated_sensitivity",
    "abs_sum_groupwise_sensitivities",
    "sum_opposite_sign_groupwise_sensitivities",
    "sensitivities",
    "uncertainties"
]
"""Names of the data fields parsed by the SDF reader for TSUNAMI-B formatted SDF files."""

COMPARISON_HEATMAP_LABELS = {
    'Calculated': 'Calculated Integral Index',
    'TSUNAMI-IP': 'TSUNAMI-IP Integral Index',
    'Percent Difference': 'Percent Difference (%)'
}
"""Labels for comparison heatmaps."""

generating_docs = False
"""Whether or not to kill interactive legend (flask/dash applications) plots after a short amount of time. This is
not intended for use by users, but is necessary for generating documentation properly."""

cache_dir = Path("~/.tsunami_ip_utils_cache").expanduser()
"""Directory to store cached cross section libraries and perturbations. This is also where the package will look for already cached
data, so be sure it corresponds with where your cached data actually is, if you have manually changed this."""


"""This module contains the functions necessary for processing the binary SCALE multigroup cross section libraries
(using the AMPX tools: charmin and tabasco) into python friendly dictionaries of numpy arrays."""

from pyparsing import *
import numpy as np
from pathlib import Path
from string import Template
import tempfile
import subprocess, os
import multiprocessing
from functools import partial
import re
from typing import Union, Tuple, Dict, Any, Callable, List
import typing
import random

def _parse_nuclide_reaction(filename: Union[str, Path], energy_boundaries: bool=False) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
    """Reads a multigroup cross section ``.txt`` file produced by the extractor function and returns the energy-dependent 
    cross sections as a numpy array.
    
    Parameters
    ----------
    filename
        The filename of the cross section file
    energy_boundaries
        If True, the energies at which the cross sections are defined are returned as well
    
    Returns
    -------
        * If ``energy_boundaries`` is ``False``, a numpy array of cross sections
        * If ``energy_boundaries`` is ``True``, a numpy array of cross sections and a numpy array of energy boundaries"""
    xs = {}
    with open(filename, 'r') as f:
        data = f.read()

    # ---------------------------
    # Define grammar for parsing
    # ---------------------------

    xs_data_line = Suppress(pyparsing_common.sci_real) + pyparsing_common.sci_real + Suppress(LineEnd())

    # Note that the output is formatted such that the same cross section value is printed for both energy boundaries of the 
    # group to avoid duplicating the cross section data, skip every other data line
    xs_parser = OneOrMore(xs_data_line + Suppress(xs_data_line))

    xs = np.array(xs_parser.parseString(data).asList())

    if energy_boundaries:
        # Define a parser that reads the energy boundaries of the groups
        energy_data_line = pyparsing_common.sci_real + Suppress(pyparsing_common.sci_real + LineEnd())
        energy_parser = OneOrMore(energy_data_line)
        energy_boundaries = np.unique(energy_parser.parseString(data).asList())
        return xs, energy_boundaries
    else:
        return xs

def _parse_reactions_from_nuclide(filename: Union[str, Path], **kwargs: dict) -> Dict[str, np.ndarray]:
    """Reads a set of reactions (given by the list of reaction mt's) from a dump of all reactions for a single nuclide from
    a SCALE library. Note this function requires that the dump included headers
    
    Parameters
    ----------
    filename
        The filename of the dump file.
    **kwargs
        Additional keyword arguments:

        - reaction_mts (List[str])
            The list of reaction MTs to read (**required** kwarg).
        - energy_boundaries (bool)
            If ``True``, the energies at which the cross sections are defined are returned as well (optional kwarg).
    
    Returns
    -------
        A dictionary containing the cross sections for each reaction MT"""

    if 'reaction_mts' not in kwargs:
        raise ValueError("Missing required keyword argument: reaction_mts")
    
    reaction_mts = kwargs['reaction_mts']
    energy_boundaries = kwargs.get('energy_boundaries', False)

    if energy_boundaries:
        raise NotImplementedError("Energy boundaries are not yet supported for this function")

    with open(filename, 'r') as f:
        data = f.read()

    # ===========================
    # Define grammar for parsing
    # ===========================

    zaid = Word(nums, max=7)
    reaction_mt = Word(nums, max=4)
    fido_field = Word(nums + '$')
    fido_subfield = Word(nums + '#')

    # -------------------------------
    # Define the header line by line
    # -------------------------------
    subfield_end = Literal('t') + LineEnd()
    other_subfield_end = Literal('e') + Literal('t') + LineEnd()

    # Define a field bundle
    bundle_line1 = Suppress(fido_field) + Suppress(zaid) + Suppress(Word(nums)) + reaction_mt
    bundle_line2 = Suppress(OneOrMore(Word(nums)))
    bundle_line3 = Suppress(fido_subfield + Word(alphanums) + OneOrMore(pyparsing_common.fnumber) + other_subfield_end)
    field_bundle = bundle_line1 + bundle_line2 + bundle_line3
    
    misc_field = fido_field + Word(nums) + Word(nums)

    header = Suppress(field_bundle) + \
             field_bundle +\
             Suppress(Optional(field_bundle)) + \
             Suppress(misc_field) + \
             Suppress(fido_subfield)

    # -------------------------------------------
    # Define the cross section data line by line
    # -------------------------------------------
    xs_data_line = Suppress(pyparsing_common.sci_real) + pyparsing_common.sci_real + Suppress(LineEnd())

    # -------------------------------------------
    # Now define the total parser for a reaction
    # -------------------------------------------
    reaction_parser = header + Group(OneOrMore(xs_data_line + Suppress(xs_data_line))) + Suppress(subfield_end)

    #--------------------------------
    # Parse the data and postprocess
    #--------------------------------
    parsed_data = reaction_parser.searchString(data)
    parsed_data = { match[0]: np.array(match[1]) for match in parsed_data }
    all_mts = parsed_data.keys()
    parsed_data = { mt: data for mt, data in parsed_data.items() if mt in reaction_mts }
    if parsed_data.keys() != set(reaction_mts):
        raise ValueError(f"Not all reaction MTs were found in the data. Missing MTs: {set(reaction_mts) - set(parsed_data.keys())}. "
                         f"This nuclide has the available MTs: {list(all_mts)}")
    return parsed_data

def _parse_from_total_library(filename: Union[str, Path], **kwargs: dict
                              ) -> Union[Dict[str, np.ndarray], Tuple[Dict[str, np.ndarray], Dict[str, list]]]:
    """Parse selected nuclide-reactions from an entire cross section library dump.
    
    Parameters
    ----------
    filename
        Path to the file containing the library dump.
    **kwargs
        Additional keyword arguments:
        
        - nuclide_reaction_dict (dict)
            A dictionary mapping nuclides to a list of reaction MTs to read (**required** kwarg).
        - return_available_nuclide_reactions (bool)
            If ``True``, the available nuclide reactions (i.e. all of the nuclide reactions for which data exists in the
            given multigroup librari) are returned as well (optional kwarg).
        - energy_boundaries (bool)
            If ``True``, the energies at which the cross sections are defined are returned as well (optional kwarg).
            
    Returns
    -------
        * If ``return_available_nuclide_reactions`` is ``False``, a doubly nested dictionary containing the cross sections for each
          nuclide-reaction pair. Keyed first by nuclide, then by reaction.
        * If ``return_available_nuclide_reactions`` is ``True``, a tuple containing the above dictionary and a dictionary containing
          all available nuclide reactions."""
    if 'nuclide_reaction_dict' not in kwargs:
        raise ValueError("Missing required keyword argument: nuclide_reaction_dict")
    
    if 'return_available_nuclide_reactions' in kwargs:
        return_available_nuclide_reactions = kwargs['return_available_nuclide_reactions']
    else:
        return_available_nuclide_reactions = False

    nuclide_reaction_dict = kwargs['nuclide_reaction_dict']
    energy_boundaries = kwargs.get('energy_boundaries', False)

    if energy_boundaries:
        raise NotImplementedError("Energy boundaries are not yet supported for this function")

    with open(filename, 'r') as f:
        data = f.read()

    # ===========================
    # Define regex patterns
    # ===========================

    header_pattern = re.compile(r'2\$\$\s+(\d+)\s+\d+\s+(\d+).*?(?=2##)', re.DOTALL)
    xs_data_pattern = re.compile(r'4##\s*((?:\s*[-+]?(?:\d*\.\d+|\d+\.?\d*)(?:[eE][-+]?\d+)?\s+[-+]?(?:\d*\.\d+|\d+\.?\d*)(?:[eE][-+]?\d+)?\s*\n)+)', re.MULTILINE)

    # ===========================
    # Parse the data
    # ===========================

    parsed_data_dict = {}
    all_nuclide_reactions = {}
    for match in header_pattern.finditer(data):
        nuclide = match.group(1)
        reaction = match.group(2)
        header_end = match.end()

        # Now record all available nuclide reactions
        if nuclide not in all_nuclide_reactions:
            all_nuclide_reactions[nuclide] = []
        
        duplicate_reaction = False
        if reaction != '0':
            if reaction in all_nuclide_reactions[nuclide]:
                duplicate_reaction = True # Not sure why these occur
            else:
                all_nuclide_reactions[nuclide].append(reaction)

        # Check if the nuclide and reaction are in the nuclide_reaction_dict
        nuclide_reaction_is_requested = nuclide in nuclide_reaction_dict and reaction in nuclide_reaction_dict[nuclide]
        if nuclide_reaction_is_requested and not duplicate_reaction:
            xs_data_match = xs_data_pattern.search(data, header_end)
            if xs_data_match:
                xs_data_text = xs_data_match.group(1)
                lines = xs_data_text.strip().split('\n')
                xs_data = [float(line.split()[1]) for line in lines[::2]]  # Extract second column and skip every other row
                if nuclide not in parsed_data_dict:
                    parsed_data_dict[nuclide] = {}
                parsed_data_dict[nuclide][reaction] = np.array(xs_data)

    # ========================================
    # Check for missing nuclides and reactions
    # ========================================

    nuclides_not_found = set(nuclide_reaction_dict.keys()) - set(parsed_data_dict.keys())
    reactions_not_found = {}
    for nuclide, reactions in nuclide_reaction_dict.items():
        # Skip the nuclides that aren't found
        if nuclide in nuclides_not_found:
            continue
        for reaction in reactions:
            if reaction not in parsed_data_dict[nuclide]:
                if nuclide not in reactions_not_found:
                    reactions_not_found[nuclide] = []
                reactions_not_found[nuclide].append(reaction)

    if len(nuclides_not_found) > 0 or reactions_not_found != {}:
        raise ValueError(f"Not all requested reactions were found in the data. Missing reactions: {reactions_not_found}. "
                         f"And missing nuclides: {nuclides_not_found}")


    if return_available_nuclide_reactions:
        # Remove the nuclide '0' from the list of available nuclides if it exists
        all_nuclide_reactions.pop('0', None)
        return parsed_data_dict, all_nuclide_reactions
    else:
        return parsed_data_dict

def _read_nuclide_reaction_from_multigroup_library(multigroup_library_path: Path, nuclide_zaid: str, reaction_mt: str, 
                                                   parsing_function: Callable[[Union[str, Path], bool, dict], Any]=_parse_nuclide_reaction, 
                                                   plot_option: str='plot', energy_boundaries: bool=False, **kwargs: dict) -> Any:
    """Uses SCALE to dump a binary multigroup library to a text file, and then calls the specified parsing function on the 
    output file.
    
    Parameters
    ----------
    multigroup_library_path
        The path to the SCALE multigroup library file.
    nuclide_zaid
        The ZAID of the nuclide.
    reaction_mt
        The reaction MT to read.
    parsing_function
        The function to call on the output file. This function takes the filename of the output file as its first argument, and
        whether or not energy boundaries should be returned as its second argument. Additional keyword arguments can be passed
        to the parsing function using the kwargs argument.
    plot_option
        The plot option to use when running the MG reader. For example ``'plot'`` or ``'fido'``.
    energy_boundaries
        If True, the energies at which the cross sections are defined are returned as well.
    
    Returns
    -------
        An output that is the result of the parsing function"""
    # Get the directory of the current file
    current_dir = Path(__file__).parent

    # Construct the path to the input file
    file_path = current_dir / 'input_files' / 'MG_reader.inp'

    # Create a tempfile for storing the output file of the MG reader dump.
    with tempfile.NamedTemporaryFile(mode='w', delete=False) as output_file:
        output_file_path = output_file.name
    
    # Read the MG reader input template file
    with open(file_path, 'r') as f:
        template = Template(f.read())
        
    # Substitute the input file template variables
    input_file = template.safe_substitute(
        nuclide_zaid=nuclide_zaid, 
        reaction_mt=reaction_mt, 
        multigroup_library_path=multigroup_library_path,
        output_file_path=output_file_path,
        plot_option=plot_option
    )

    # Write the input file to another tempfile
    with tempfile.NamedTemporaryFile(mode='w', delete=False) as input_temp_file:
        input_temp_file.write(input_file)
        input_temp_file_path = input_temp_file.name

    # Run the executable
    command = ['scalerte', input_temp_file_path]

    proc = subprocess.Popen(command)
    proc.wait()

    # Now delete the input file
    os.remove(input_temp_file_path)

    # Read the output file
    output = parsing_function(output_file_path, energy_boundaries=energy_boundaries, **kwargs)

    # Now delete the output file
    os.remove(output_file_path)
    return output

def _read_reactions_from_nuclide(multigroup_library_path: Path, nuclide_zaid: str, reaction_mts: List[str]
                                 ) -> Dict[str, np.ndarray]:
    """Function for reading a set of reactions from a given nuclide in a SCALE multigroup library.
    
    Parameters
    ----------
    multigroup_library_path
        The path to the SCALE multigroup library file.
    nuclide_zaid
        The ZAID of the nuclide.
    reaction_mts
        The list of reaction MTs to read.
    
    Returns
    -------
        A dictionary containing the cross sections for each reaction MT"""
    output = _read_nuclide_reaction_from_multigroup_library(multigroup_library_path, nuclide_zaid, reaction_mt='0', \
                                                           parsing_function=_parse_reactions_from_nuclide, \
                                                            reaction_mts=reaction_mts, plot_option='fido')
    return output

def read_multigroup_xs(multigroup_library_path: Path, nuclide_zaid_reaction_dict: Dict[str, Dict[str, str]], 
                       num_processes: int=multiprocessing.cpu_count(), return_available_nuclide_reactions: bool=False
                       ) -> Union[Dict[str, Dict[str, np.ndarray]], Tuple[Dict[str, Dict[str, np.ndarray]], Dict[str, list]]]:
    """Function for reading a set of reactions from a given nuclide in a SCALE multigroup library.
    
    Parameters
    ----------
    multigroup_library_path
        The path to the SCALE multigroup library file.
    nuclide_zaid_reaction_dict
        A dictionary mapping nuclide ZAIDs to a list of reaction MTs to read.
    num_processes
        The number of processes to use for reading the library. If None, the number of processes is set to the number of cores.
    return_available_nuclide_reactions
        If True, the available nuclide reactions (i.e. all of the nuclide reactions for which data exists in the given 
        multigroup library) are returned as well.

    Returns
    -------
        * If ``return_available_nuclide_reactions`` is ``False``, a dictionary containing the cross sections (as numpy arrays) 
          for each nuclide-reaction pair. Keyed first by nuclide, then by reaction.
        * If ``return_available_nuclide_reactions`` is ``True``, a tuple containing the above dictionary and a 
          dictionary containing all available nuclide reactions is returned.
    """

    NUCLIDE_THRESHOLD = 50 # Number of nuclides after which the large method is more performant and hence is used
    CORE_THRESHOLD = 2 # The large method is more performant if the number of cores is smaller than this number

    num_nuclides = len(list(nuclide_zaid_reaction_dict.keys()))
    use_small_method = ( num_nuclides < NUCLIDE_THRESHOLD ) and ( num_processes >= CORE_THRESHOLD )
    if use_small_method and not return_available_nuclide_reactions: # This method is slow but works for small amounts of nuclide reactions or a large amount of cores
        pool = multiprocessing.Pool(processes=num_processes)

        # Create a partial function with the common arguments
        read_reactions_partial = partial(_read_reactions_from_nuclide, multigroup_library_path)

        # Distribute the function calls among the processes
        results = pool.starmap(read_reactions_partial, nuclide_zaid_reaction_dict.items())

        # Close the pool and wait for the processes to finish
        pool.close()
        pool.join()

        # Convert the results to a dictionary
        output = dict(zip(nuclide_zaid_reaction_dict.keys(), results))

        return output
    else: # This method is faster (as in there's less scale run overhead) but requires the entire library to be read and is serial
        # If the user wants the available nuclide reactions, then we need to create a partial function which adds the appropriate
        # keyword argument to the parsing function
        if return_available_nuclide_reactions:
            parse_function = partial(_parse_from_total_library, return_available_nuclide_reactions=True)
        else:
            parse_function = _parse_from_total_library
        
        output = _read_nuclide_reaction_from_multigroup_library(multigroup_library_path, nuclide_zaid='0', reaction_mt='0', \
                                                        parsing_function=parse_function, \
                                                        nuclide_reaction_dict=nuclide_zaid_reaction_dict, plot_option='fido')
        return output
    
def perturb_multigroup_xs_dump(filename: Union[str, Path],  max_perturb_factor: float, overwrite: bool=False,
                               output_file: typing.Optional[Union[str, Path]]=None) -> typing.Optional[List[str]]:
    """Perturb the cross section data in a SCALE multigroup cross section library fido text dump file. This is useful for generating
    examples for testing the SCALE reader functions which do not violate export control.
    
    Parameters
    ----------
    filename
        The filename of the fido text dump file.
    max_perturb_factor
        The maximum percentage by which to perturb the cross sections. Cross sections are perturbed by a random factor between
        ``1 - max_perturb_factor`` and ``1 + max_perturb_factor``.
    overwrite
        Whether or not to overwrite the file with the perturbed data.
    output_file
        An output file to write the perturbed data to.
        
    Returns
    -------
        * If ``True``, the file is overwritten with the perturbed data and nothing is returned.
        * If ``False``, the perturbed data is returned as a list of strings, which can be written to a file via
          ``with open('filename.txt', 'w') as f: f.writelines(perturbed_data)``.
          
    Notes
    -----
    - This only applies to `fido` dumps of SCALE multigroup cross section libraries.
    - Fido dumps of SCALE multigroup cross section libraries can be generated using """
    # First read the multigroup xs library text dump
    with open(filename, 'r') as f:
        input_file = f.readlines()

    # Perturb the data
    reading_data = False
    perturbed_data = []
    for line in input_file:
        # Identify the start of the data with the correct fido delimiters
        if line.startswith(' 4## '):
            reading_data = True
            last_energy = None
            newline = line
            perturbed_data.append(newline)
            continue

        if reading_data and line.startswith(' t'):
            reading_data = False

        # Now perturb the data
        if reading_data:
            energy, xs = line.strip().split()
            if energy == last_energy:
                # If the energy is the same as the last energy, keep the (perturbed) xs the same
                newline = f'        {energy} {last_xs}\n'
                last_xs = last_xs
            else:
                perturbation_factor = 1.0 + random.uniform(-max_perturb_factor, max_perturb_factor)
                newline = f'        {energy} {float(xs) * perturbation_factor}\n'
                last_xs = float(xs) * perturbation_factor
            last_energy = energy
        else:
            newline = line

        perturbed_data.append(newline)
    
    if overwrite:
        with open(filename, 'w') as f:
            f.writelines(perturbed_data)
    elif output_file is not None:
        with open(output_file, 'w') as f:
            f.writelines(perturbed_data)
    else:
        return perturbed_data


"""Tools for creating interactive and static scatter plots with error bars, linear regression lines, and correlation coefficient
calculations."""

from __future__ import annotations
import scipy.stats as stats
import numpy as np
import matplotlib.pyplot as plt
from ._base_plotter import _Plotter
import plotly.express as px
import plotly.graph_objects as go
from plotly.subplots import make_subplots
import pandas as pd
import plotly.graph_objects as go
import dash
from dash import dcc, html
from dash.dependencies import Input, Output, State
import pandas as pd
import webbrowser
import os
import signal
import pickle
import threading
from .plot_utils import _find_free_port
import sys, os, signal
import threading
import webbrowser
import sys
import plotly.graph_objs as go
from pyparsing import *
import re
from uncertainties import ufloat
import tsunami_ip_utils
from typing import Tuple, Any, Union, List
import typing
from matplotlib.figure import Figure
from matplotlib.axes import Axes
from pathlib import Path
import plotly.io as pio
import tempfile
from tsunami_ip_utils.viz.plot_utils import _capture_html_as_image
import tsunami_ip_utils.config as config
from matplotlib.patches import Rectangle

def _replace_spearman_and_pearson(text: str, new_pearson: float, new_spearman: float) -> str:
    """Replaces the Spearman and Pearson values in the given text with the new values provided using regex. This is useful
    for updating the correlation statistics in a scatter plot annotation interactively after the plot has been created.
    
    Parameters
    ----------
    text
        The text (scatter plot annotation) containing the Spearman and Pearson values to be updated. It is expected that the
        text contains the Spearman and Pearson values in the format ``"Spearman: <b>0.123456</b> Pearson: <b>0.123456</b>"``.
    new_pearson
        The new Pearson correlation coefficient value to replace the old value in the text.
    new_spearman
        The new Spearman rank correlation coefficient value to replace the old value in the text.

    Returns
    -------
        The updated text with the new Spearman and Pearson values."""
    
    new_pearson = f"{new_pearson:1.6f}"
    new_spearman = f"{new_spearman:1.6f}"

    # Replace Pearson value using regex
    pearson_pattern = r"(Pearson: <b>)[0-9.]+(</b>)"
    text = re.sub(pearson_pattern, r"\g<1>" + new_pearson + r"\2", text)

    # Replace Spearman value using regex
    spearman_pattern = r"(Spearman: <b>)[0-9.]+(</b>)"
    text = re.sub(spearman_pattern, r"\g<1>" + new_spearman + r"\2", text)
    
    return text


def _update_percent_difference(text: str, reference_value: float) -> Tuple[str, ufloat]:
    """Updates the percent difference in the given text with the new percent difference calculated using the new Pearson
    correlation coefficient value. This is useful for updating the percent difference in a scatter plot annotation interactively
    after the plot has been created. This is only valid for scatter plots generated by 
    :func:`tsunami_ip_utils.comparisons.correlation_comparison` or other plots that have the percent difference in the annotation.
    
    Parameters
    ----------
    text
        The text (scatter plot annotation) containing the percent difference to be updated. It is expected that the text
        contains the percent difference in the format ``"Percent Difference: <b>0.123456</b>%"`` and the TSUNAMI-IP c_k value
        in the format ``"TSUNAMI-IP c_k: <b>0.123456+/-0.123456</b>"``.
    reference_value
        The reference value to calculate the percent difference from. This is typically the (updated) Pearson correlation 
        coefficient.
    
    Returns
    -------
        - updated_text
            The updated text with the new percent difference value.
        - percent_difference
            The new percent difference value with uncertainty."""
    # Regex to find the TSUNAMI-IP c_k value
    ck_pattern = r"TSUNAMI-IP c_k: <b>([\d\.]+)\+/-([\d\.]+)</b>"
    # Find the TSUNAMI-IP c_k value in the text
    match = re.search(ck_pattern, text)
    if not match:
        raise ValueError("TSUNAMI-IP c_k value not found in the text")
    
    # Parse the nominal value and the uncertainty
    nominal_value, uncertainty = match.groups()
    tsunami_ck = ufloat(float(nominal_value), float(uncertainty))
    
    # Calculate the percent difference
    if reference_value == 0 or np.isnan(reference_value):
        percent_difference = ufloat(0, 0)
    else:
        percent_difference = (tsunami_ck - reference_value) / reference_value * 100
    
    # Format the new percent difference with uncertainty
    new_percent_diff_text = f"<b>{percent_difference.nominal_value:.2f}+/-{percent_difference.std_dev:.2f}</b>%"
    
    # Regex to replace the Percent Difference in the text
    # Escaping special characters used in the regex pattern
    updated_text = re.sub(r"Percent Difference: <b>[\d\.\+\-/]+</b>+%", f"Percent Difference: {new_percent_diff_text}", text)
    
    return updated_text, percent_difference

class EnhancedPlotlyFigure(go.Figure):
    """This class wraps a plotly express figure object (intended for a scatter plot) and adds additional attributes for the
    summary statistics and linear regression data. This class is intended to be used with the :class:`.InteractiveScatterPlotter`
    class."""
    statistics: dict
    """A dictionary containing the Pearson and Spearman correlation coefficients."""
    regression: dict
    """A dictionary containing the slope and intercept of the linear regression line."""
    def __init__(self, *args, **kwargs):
        """Initializes an EnhancedPlotlyFigure object from a Plotly Express figure object.
        
        Intended Use
        ============
        This class should be initialized from a plotly expres Figure object via:
        
        .. code-block:: python
            
            fig = px.scatter(...)
            fig = EnhancedPlotlyFigure(fig.to_dict())
            
        Additional ``statistics`` and ``regression`` attributions can be added to this "enhanced" figure via
        
        .. code-block:: python
        
            fig.statistics = {
                'pearson': 0.123456,
                'spearman': 0.123456
            }
            fig.regression = {
                'slope': 0.123456,
                'intercept': 0.123456
            }
        
        unlike a regular Plotly Express figure object, which will throw an error if you try to add these attributes directly."""
        super().__init__(*args, **kwargs)

        # Directly set the attributes using object's __setattr__ to bypass Plotly's checks
        object.__setattr__(self, 'statistics', None)
        object.__setattr__(self, 'regression', None)

    def __setattr__(self, name, value):
        """Sets the attribute of the object. This method is overridden to allow setting the ``'statistics'`` and ``'regression'``
        attributes directly without raising an error. All other attributes are set using the super class's __setattr__ method."""
        if name in ['statistics', 'regression']:
            # Handle custom attributes internally
            object.__setattr__(self, name, value)
        else:
            # Use the super class's __setattr__ for all other attributes
            super().__setattr__(name, value)

class _ScatterPlot(_Plotter):
    """This class exists to add some additional functionality for calculating regressions and summary statistics that's
    common to all types of scatter plots, interactive or otherwise"""
    _regression: Any
    """A named tuple (``scipy.stats`` ``Linregress`` object) containing the slope and intercept of the linear regression line
    as attributes ``slope`` and ``intercept`` respectively."""
    _pearson: float
    """The Pearson correlation coefficient."""
    _spearman: float
    """The Spearman rank correlation coefficient."""
    _summary_stats_text: str
    """A string containing the summary statistics for the scatter plot."""
    def _get_summary_statistics(self, x: Union[List, np.ndarray], y: Union[List, np.ndarray]) -> None:
        """Calculates the Pearson correlation coefficient, Spearman rank correlation coefficient, and linear regression
        parameters for the given x and y datasets. The linear regression parameters are the slope and intercept of the
        regression line. The Pearson and Spearman coefficients are also stored in the class instance as 'pearson' and
        'spearman' respectively. The slope and intercept are stored as 'slope' and 'intercept' respectively. The linear
        regression is stored as 'regression'
        
        Parameters
        ----------
        x
            The x values of the scatter plot.
        y
            The y values of the scatter plot."""
        self._regression = stats.linregress(x, y)
        self._pearson    = stats.pearsonr(x, y).statistic
        self._spearman   = stats.spearmanr(x, y).statistic
        self._slope      = self._regression.slope
        self._intercept  = self._regression.intercept

        # If the figure has been plotted (and is an enhanced plot which supports adding this metadata), add the regression 
        # and correlation statistics to the figure
        is_enhanced_fig = isinstance(getattr(self, 'fig', None), EnhancedPlotlyFigure)
        if hasattr(self, 'fig') and is_enhanced_fig:
            self.fig.statistics = {
                'pearson': self._pearson,
                'spearman': self._spearman
            }
            self.fig.regression = {
                'slope': self._slope,
                'intercept': self._intercept,
            }

        # Now create the summary statistics text for figure annotation
        if self._plot_type == 'plotly':
            pearson_text = f"Pearson: <b>{self._pearson:1.6f}</b>"
            spearman_text = f"Spearman: <b>{self._spearman:1.6f}</b>"
        elif self._plot_type == 'matplotlib':
            pearson_text = f"Pearson: {self._pearson:1.6f}"
            spearman_text = f"Spearman: {self._spearman:1.6f}"
        else:
            raise ValueError(("Unknown plot type, cannot format summary statistics text"))
        self._summary_stats_text = f"{pearson_text} {spearman_text}"


class _ScatterPlotter(_ScatterPlot):
    """A class for creating static scatter plots with error bars, linear regression lines, and correlation coefficient
    calculations."""
    _nested: bool
    """Whether the contributions are nested or not. This is unused in this class."""
    _index_name: str
    """The name of the integral index (whose contributions) being plotted. This is used only for the title of the plot."""
    _plot_redundant: bool
    """Whether to plot redundant reactions or not. This is unused in this class."""
    _plot_type: str
    """Whether the plot is a matplolib or plotly plot. This is used to determine how to format the summary statistics text."""
    def __init__(self, integral_index_name: str, nested: bool, plot_redundant: bool=False, **kwargs: dict) -> None:
        """Initializes a ScatterPlotter object with the given integral index name, nested status, and plot redundant reactions
        status. The nested and plot_redundant arguments are unused in this class.
        
        Parameters
        ----------
        integral_index_name
            The name of the integral index (whose contributions) being plotted.
        nested
            Whether the contributions are nested or not.
        plot_redundant
            Whether to plot redundant reactions or not.
        kwargs
            Additional keyword arguments. These are unused in this class."""
        self._nested = nested
        self._index_name = integral_index_name
        self._plot_redundant = plot_redundant
        self._plot_type = 'matplotlib'

    def _create_plot(self, contribution_pairs: List[ufloat], isotopes: List[str], reactions: List[str]) -> None:
        """Creates a static scatter plot with error bars, linear regression line, and correlation coefficient calculations.
        
        Parameters
        ----------
        contribution_pairs
            A list of pairs of contributions to the integral index from the application and experiment.
        isotopes
            The list of isotopes represented by each contribution pair. This has the same length and it ordered the same
            as the contribution_pairs.
        reactions
            The list of all reactions. For consistency, this is the same for all isotopes, so there may be cases where an
            isotope has a reaction that isn't actually possible (e.g. c-12 fission) but is included for consistency with the
            expeirment/application, and the contribution is zero. 
        
        Notes
        -----
        The set of isotopes and reactions is only necessary for creating labels in interactive plots (e.g. those made by 
        :class:`.InteractiveScatterPlotter`), but is included here for consistency."""
        self.fig, self.axs = plt.subplots(figsize=(8, 8))

        # Extract the x and y values from the contribution pairs
        application_points        = [ contribution[0].n for contribution in contribution_pairs ]
        application_uncertainties = [ contribution[0].s for contribution in contribution_pairs ]
        experiment_points         = [ contribution[1].n for contribution in contribution_pairs ]
        experiment_uncertainties  = [ contribution[1].s for contribution in contribution_pairs ]

        print(len(application_points), len(isotopes)*len(reactions))

        # Now filter out (0,0) points, which don't contribute to either the application or the experiment, these are
        # usually chi, nubar, or fission reactions for nonfissile isotopes that are added for consistency with the set
        # of reactions only

        indices = [ index for index, (application_point, experiment_point) in enumerate( zip( application_points, experiment_points ) )
                    if application_point == 0 and experiment_point == 0 ]
        
        application_points        = [ point for index, point in enumerate(application_points) if index not in indices ]
        application_uncertainties = [ point for index, point in enumerate(application_uncertainties) if index not in indices ]
        experiment_points         = [ point for index, point in enumerate(experiment_points) if index not in indices ]
        experiment_uncertainties  = [ point for index, point in enumerate(experiment_uncertainties) if index not in indices ]
        isotopes                  = [ isotope for index, isotope in enumerate(isotopes) if index not in indices ]

        # # Calculate the magnitude (absolute value of x ordinate plus absolute value of y ordinate)
        # magnitudes = [abs(application_points[i]) + abs(experiment_points[i]) for i in range(len(application_points))]

        # # Combine all lists to sort them together
        # combined = list(zip(magnitudes, application_points, application_uncertainties, experiment_points, experiment_uncertainties, isotopes))

        # # Sort the combined list by magnitudes
        # combined.sort(key=lambda x: x[0], reverse=True)

        # # Unzip the sorted lists
        # magnitudes, application_points, application_uncertainties, experiment_points, experiment_uncertainties, isotopes = zip(*combined)

        # # Define a color map
        colors = plt.cm.rainbow(np.linspace(0, 1, len(application_points)))

        # Plot each point with error bars and color by order of magnitude
        for i, (x, xerr, y, yerr, color, isotope) in enumerate(zip(application_points, application_uncertainties, experiment_points, experiment_uncertainties, colors, isotopes)):
            if reactions != []:
                for reaction in reactions:
                    self.axs.errorbar(x, y, xerr=xerr, yerr=yerr, fmt='.', capsize=5, color=color, label=f"{isotope} : {reaction}" )
            else:
                self.axs.errorbar(x, y, xerr=xerr, yerr=yerr, fmt='.', capsize=5, color=color, label=f"{isotope}")

        self.axs.errorbar(application_points, experiment_points, xerr=application_uncertainties, \
            yerr=experiment_uncertainties, fmt='.', capsize=5)
        # Linear regression
        self._get_summary_statistics(application_points, experiment_points)

        # Plot the regression line
        x = np.linspace(min(application_points), max(application_points), 100)
        y = self._slope * x + self._intercept
        self.axs.plot(x, y, 'r', label='Linear fit')

        self.axs.text(0.05, 0.95, self._summary_stats_text, transform=self.axs.transAxes, fontsize=12,
                verticalalignment='top', bbox=dict(facecolor='white', alpha=0.5))

        self._style()

    def _add_legend_with_scaling(self):
        def update_legend(fontsize):
            legend = self.axs.legend(bbox_to_anchor=(1, 1), loc='upper left')
            for text in legend.get_texts():
                text.set_fontsize(fontsize)
            return legend

        def legend_fits():
            fig_width, fig_height = self.fig.get_size_inches()
            legend = self.axs.get_legend()
            if legend:
                # Estimate the width of the legend
                renderer = self.fig.canvas.get_renderer()
                bbox = legend.get_window_extent(renderer)
                legend_width = bbox.width / self.fig.dpi
                legend_height = bbox.height / self.fig.dpi
                width_fits = fig_width - legend_width > 0.1  # Allow a margin for axes labels and other elements
                height_fits = fig_height - legend_height > 0.1  # Allow a margin for axes labels and other elements
                return width_fits and height_fits
            return True

        # Initial legend attempt
        legend = update_legend(10)

        # Try to fit the legend using direct size check
        if not legend_fits():
            # If the legend doesn't fit, scale the font size
            for fontsize in range(10, 5, -1):
                legend = update_legend(fontsize)
                if legend_fits():
                    break

            # If still doesn't fit, exclude the smallest magnitudes
            if not legend_fits():
                handles, labels = self.axs.get_legend_handles_labels()
                num_to_exclude = 1
                while not legend_fits() and num_to_exclude < len(labels):
                    num_to_exclude += 1
                    self.axs.legend(handles[:-num_to_exclude], labels[:-num_to_exclude], bbox_to_anchor=(1, 1), loc='upper left')

                # Exclude an additional legend entry to make room for the annotation
                num_to_exclude += 2
                extra = Rectangle((0, 0), 1, 1, fc="w", fill=False, edgecolor='none', linewidth=0)
                self.axs.legend(handles[:-num_to_exclude] + [extra], labels[:-num_to_exclude] + ['others not shown'], bbox_to_anchor=(1, 1), loc='upper left')

    def _get_plot(self) -> Tuple[Figure, Axes]:
        return self.fig, self.axs
        
    def _add_to_subplot(self, fig, position) -> Figure:
        return fig.add_subplot(position, sharex=self.axs, sharey=self.axs)
    
    def _style(self):
        title_text = f'Contributions to {self._index_name}'
        self.axs.set_title(title_text)
        self.axs.set_ylabel(f"Experiment {self._index_name} Contribution")
        self.axs.set_xlabel(f"Application {self._index_name} Contribution")
        self.axs.grid()
        # self._add_legend_with_scaling()
        self.fig.tight_layout()

def load_interactive_scatter_plot(filename: Union[str, Path]) -> InteractiveScatterLegend:
    """Loads an interactive scatter plot (with interactive legend) from a saved state file. This function is purely for 
    convenience and is a wrapper of the :meth:`.InteractiveScatterLegend.load_state` method
    
    Parameters
    ----------
    filename
        The filename of the saved state file to load the interactive scatter plot from.

    Returns
    -------
        A reseralized instance of the interactive scatter plot that can be plotted with :meth:`InteractiveScatterLegend.show()`."""
    
    return InteractiveScatterLegend.load_state(filename)


class _InteractiveScatterPlotter(_ScatterPlot):
    """A class for creating `interactive` scatter plots with error bars, linear regression lines, and correlation coefficient
    calculations."""
    _interactive_legend: bool
    """Whether to include an interactive legend in the plot. This is used to toggle the visibility of the traces in the plot
    and interactively recalculate the regression and summary statistics."""
    _nested: bool
    """Whether the contributions are nested or not. This is unused in this class."""
    _index_name: str
    """The name of the integral index (whose contributions) being plotted. This is used only for the title of the plot."""
    _plot_redundant: bool
    """Whether to plot redundant reactions or not. This is unused in this class."""
    _plot_type: str
    """Whether the plot is a matplolib or plotly plot. This is used to determine how to format the summary statistics text."""
    def __init__(self, integral_index_name: str, nested: bool, plot_redundant: bool=False, **kwargs: dict):
        """Initializes an ``InteractiveScatterPlotter`` object with the given options.
        
        Parameters
        ----------
        integral_index_name
            The name of the integral index (whose contributions) being plotted.
        nested
            Whether the contributions are nested or not.
        plot_redundant
            Whether to plot redundant reactions or not.
        kwargs
            Additional keyword arguments.

            - interactive_legend (bool)
                Whether to include an interactive legend in the plot. Default is ``False``."""
        if 'interactive_legend' in kwargs.keys():
            self._interactive_legend = kwargs['interactive_legend']
        else:
            self._interactive_legend = False
        self._nested = nested
        self._index_name = integral_index_name
        self._plot_redundant = plot_redundant
        self._plot_type = 'plotly'

    def _create_plot(self, contribution_pairs: List[ufloat], isotopes: List[str], reactions: List[str]) -> None:
        """Create an interactive scatter plot with error bars, linear regression line, and correlation coefficient calculations.
        
        Parameters
        ----------
        contribution_pairs
            A list of pairs of contributions to the integral index from the application and experiment.
        isotopes
            The list of isotopes represented by each contribution pair. This has the same length and it ordered the same
            as the ``contribution_pairs``.
        reactions
            The list of reactions represented by each contribution pair. This has the same length and it ordered the same
            as the ``contribution_pairs``."""
        self.fig = make_subplots()

        # Extract isotope and reaction pairs from the given list of isotopes and reactions
        df = self._create_scatter_data(contribution_pairs, isotopes, reactions)

        hover_data_dict = {
            'Isotope': True  # Always include Isotope
        }

        if 'Reaction' in df.columns:
            hover_data_dict['Reaction'] = True  # Include Reaction only if it exists

        # Create scatter plot with error bars using Plotly Express
        self.fig = px.scatter(
            df, 
            x=f'Application {self._index_name} Contribution', 
            y=f'Experiment {self._index_name} Contribution',
            error_x='Application Uncertainty', 
            error_y='Experiment Uncertainty',
            color='Isotope',
            labels={
                "color": "Isotope"
            },
            title=f'Contributions to {self._index_name}',
            hover_data=hover_data_dict
        )

        # Wrap the plotly express figure in a MetadataPlotly object
        self.fig = EnhancedPlotlyFigure(self.fig.to_dict())

        self._add_regression_and_stats(df)

        # Now style the plot
        self._style()

        if self._interactive_legend:
            self.fig = InteractiveScatterLegend(self, df)

    def _add_regression_and_stats(self, df: pd.DataFrame) -> None:
        """Add the linear regression line and correlation statistics to the plot. This method is called after the plot has been
        created. This method can also update a comparison with a TSUNAMI-IP calculated :math:`c_k` value if it exists in the
        annotation.
        
        Parameters
        ----------
        df
            The DataFrame containing the data for the scatter plot. This DataFrame should contain the columns ``'Application
            <integral_index_name> Contribution'`` and ``'Experiment <integral_index_name> Contribution'``.
        """
        # Calculate the linear regression and correlation statistics
        self._get_summary_statistics(df[f'Application {self._index_name} Contribution'], \
                                    df[f'Experiment {self._index_name} Contribution'])

        # Prepare data for the regression line
        x_reg = np.linspace(df[f'Application {self._index_name} Contribution'].min(), 
                            df[f'Application {self._index_name} Contribution'].max(), 100)
        y_reg = self._slope * x_reg + self._intercept

        # Convert self.fig.data to a list for mutability
        current_traces = list(self.fig.data)

        # Remove existing regression line if it exists
        traces_to_keep = [trace for trace in current_traces if not trace.name.startswith('Regression Line')]

        # Set the modified list of traces back to the figure
        self.fig.data = tuple(traces_to_keep)

        # Add new linear regression to the plot
        self.fig.add_trace(go.Scatter(x=x_reg, y=y_reg, mode='lines', 
                                    name=f'Regression Line y={self._slope:1.4E}x + {self._intercept:1.4E}'))

        # Remove existing annotation if it exists
        if hasattr(self.fig, 'layout') and hasattr(self.fig.layout, 'annotations'):
            if len(self.fig.layout.annotations) != 0:
                # Get the text of the old annotation
                old_annotation_text = \
                    [ann for ann in self.fig.layout.annotations if ann.text.startswith('Pearson')][0].text
                annotation_text = _replace_spearman_and_pearson(old_annotation_text, self._pearson, self._spearman)
                bordercolor = '#444'

                self.fig.layout.annotations = [ann for ann in self.fig.layout.annotations if not ann.text.startswith('Pearson')]
                if "TSUNAMI-IP" in annotation_text:
                    try:
                        annotation_text, percent_difference = _update_percent_difference(annotation_text, self._pearson)
                    except Exception as e:
                        print(f"Error updating percent difference: {e}")
                    if abs(percent_difference.nominal_value) > 5:
                        bordercolor = 'red'
            else:
                bordercolor = '#444'
                annotation_text = self._summary_stats_text
        else:
            bordercolor = '#444'
            annotation_text = self._summary_stats_text

        # Add correlation statistics to the plot
        self.fig.add_annotation(
            x=0.05, xref="paper", 
            y=0.95, yref="paper",
            text=annotation_text,
            bordercolor=bordercolor,
            showarrow=False, 
            font=dict(size=12),
            align='left',
            bgcolor="white", 
            opacity=0.8
        )

    def _create_scatter_data(self, contribution_pairs: List[ufloat], isotopes: List[str], reactions: List[str]) -> pd.DataFrame:
        """Create a DataFrame from the given contribution pairs, isotopes, and reactions. This DataFrame is used to create the
        interactive scatter plot.
        
        Parameters
        ----------
        contribution_pairs
            A list of pairs of contributions to the integral index from the application and experiment.
        isotopes
            The list of isotopes represented by each contribution pair. This has the same length and it ordered the same
            as the ``contribution_pairs``.
        reactions
            The list of reactions represented by each contribution pair. This has the same length and it ordered the same
            as the ``contribution_pairs``."""

        data = {
            f'Application {self._index_name} Contribution': [cp[0].n for cp in contribution_pairs],
            f'Experiment {self._index_name} Contribution': [cp[1].n for cp in contribution_pairs],
            'Application Uncertainty': [cp[0].s for cp in contribution_pairs],
            'Experiment Uncertainty': [cp[1].s for cp in contribution_pairs],
            'Isotope': [],
        }

        # Add nuclides and reactions (if they exist) to the data dictionary
        if reactions == []:
            for isotope in isotopes:
                data['Isotope'].append(isotope)
        else:
            data['Reaction'] = []
            for isotope in isotopes:
                for reaction in reactions:
                    data['Isotope'].append(isotope)
                    data['Reaction'].append(reaction)

        # Now filter out (0,0) points, which don't contribute to either the application or the experiment, these are
        # usually chi, nubar, or fission reactions for nonfissile isotopes that are added for consistency with the set
        # of reactions only
        data = { key: [val for val, app, exp in zip(data[key], data[f'Application {self._index_name} Contribution'], \
                    data[f'Experiment {self._index_name} Contribution']) if app != 0 or exp != 0] for key in data }

        return pd.DataFrame(data)

    def _add_to_subplot(self, fig, position):
        for trace in self.fig.data:
            fig.add_trace(trace, row=position[0], col=position[1])

        # Transfer annotations
        if hasattr(self.fig, 'layout') and hasattr(self.fig.layout, 'annotations'):
            for ann in self.fig.layout.annotations:
                # Adjust annotation references to new subplot
                new_ann = ann.update(xref=f'x{position[1]}', yref=f'y{position[1]}')
                fig.add_annotation(new_ann, row=position[0], col=position[1])
        return fig

    def _get_plot(self):
        return self.fig
    
    def _style(self):
        title_text = f'Contributions to {self._index_name}'
        self.fig.update_layout(title_text=title_text, title_x=0.5)  # 'title_x=0.5' centers the title


class _InteractivePerturbationScatterPlotter(_ScatterPlot):
    """Class for creating an interactive scatter plot using the nuclear data sampling method (where perturbed cross section
    libraries are used to calculate sample points on the scatter plot)."""
    _plot_type: str
    """Whether the plot is a matplolib or plotly plot. This is used to determine how to format the summary statistics text."""
    def __init__(self, **kwargs: dict):
        self._plot_type = 'plotly'
        pass

    def _create_plot(self, points: List[Tuple[ufloat, ufloat]]) -> None:
        """Create an interactive perturbation scatter plot with error bars, linear regression line, and correlation coefficient.
        
        Parameters
        ----------
        points
            A list of perturbation points, each computed from sampled perturbed cross section libraries. These points are generated
            using the :func:`tsunami_ip_utils.perturbations.generate_points` function."""
        self.fig = make_subplots()

        # Extract isotope and reaction pairs from the given list of isotopes and reactions
        df = pd.DataFrame({
            'Application': [point[0].n for point in points],
            'Experiment': [point[1].n for point in points],
            'Application Uncertainty': [point[0].s for point in points],
            'Experiment Uncertainty': [point[1].s for point in points]
        })

        # Create scatter plot with error bars using Plotly Express
        self.fig = px.scatter(
            df, 
            x=f'Application', 
            y=f'Experiment',
            error_x='Application Uncertainty', 
            error_y='Experiment Uncertainty',
            title=f'Correlation Plot',
        )

        # Wrap the plotly express figure in a MetadataPlotly object
        self.fig = EnhancedPlotlyFigure(self.fig.to_dict())

        self._add_regression_and_stats(df)

        # Now style the plot
        self._style()

    def _add_regression_and_stats(self, df: pd.DataFrame) -> None:
        """Add the linear regression line and correlation statistics to the plot. This method is called after the plot has been
        created.
        
        Parameters
        ----------
        df
            The DataFrame containing the data for the scatter plot. This DataFrame should contain the columns ``'Application'``
            and ``'Experiment'``."""
        # Calculate the linear regression and correlation statistics
        self._get_summary_statistics(df[f'Application'], df[f'Experiment'])

        # Prepare data for the regression line
        x_reg = np.linspace(df[f'Application'].min(), df[f'Application'].max(), 100)
        y_reg = self._slope * x_reg + self._intercept

        # Convert self.fig.data to a list for mutability
        current_traces = list(self.fig.data)

        # Remove existing regression line if it exists
        traces_to_keep = [trace for trace in current_traces if not trace.name.startswith('Regression Line')]

        # Set the modified list of traces back to the figure
        self.fig.data = tuple(traces_to_keep)

        # Add new linear regression to the plot
        self.fig.add_trace(go.Scatter(x=x_reg, y=y_reg, mode='lines', 
                                    name=f'Regression Line y={self._slope:1.4E}x + {self._intercept:1.4E}'))

        # Add correlation statistics to the plot
        self.fig.add_annotation(
            x=0.05, xref="paper", 
            y=0.95, yref="paper",
            text=self._summary_stats_text,
            showarrow=False,
            align='left',
            font=dict(size=12),
            bgcolor="white", 
            opacity=0.8
        )

    def _add_to_subplot(self, fig, position):
        for trace in self.fig.data:
            fig.add_trace(trace, row=position[0], col=position[1])
        return fig

    def _get_plot(self):
        return self.fig
    
    def _style(self):
        pass


class InteractiveScatterLegend(_InteractiveScatterPlotter):
    """An implementation of an interactive legend (that automatically updates the regression and summary statistics following
    the exclusion of certain data in the scatter plot) for an interactive (Plotly) scatter plot."""
    fig: EnhancedPlotlyFigure
    """The Plotly figure object containing the interactive scatter plot."""
    _interactive_scatter_plot: _InteractiveScatterPlotter
    """The interactive scatter plot object that the interactive legend is associated with."""
    df: pd.DataFrame
    """The DataFrame containing the data for the scatter plot."""
    _excluded_isotopes: List[str]
    """A list of isotopes that have been excluded from the scatter plot."""
    _app: dash.Dash
    """The Dash application object for the interactive legend."""
    _plot_type: str
    """Whether the plot is a matplolib or plotly plot. This is used to determine how to format the summary statistics text."""
    def __init__(self, interactive_scatter_plot: _InteractiveScatterPlotter, df: pd.DataFrame):
        self._interactive_scatter_plot = interactive_scatter_plot
        self.fig = interactive_scatter_plot.fig
        self._index_name = interactive_scatter_plot._index_name
        self.df = df
        self._excluded_isotopes = []  # Keep track of excluded isotopes
        self._app = dash.Dash(__name__)
        self._app.layout = html.Div([
            dcc.Graph(id='interactive-scatter', figure=self.fig, style={'height': '100vh'})
        ], style={'margin': 0})
        self._setup_callbacks()
        self._plot_type = 'plotly'

    def _setup_callbacks(self):
        """Set up the Dash callbacks for the interactive legend."""
        @self._app.callback(
            Output('interactive-scatter', 'figure'),
            Input('interactive-scatter', 'restyleData'),
            State('interactive-scatter', 'figure')
        )
        def update_figure_on_legend_click(restyleData, current_figure_state):
            if restyleData and 'visible' in restyleData[0]:
                current_fig = go.Figure(current_figure_state)

                # Get the index of the clicked trace
                clicked_trace_index = restyleData[1][0]

                # Get the name of the clicked trace
                clicked_trace_name = current_fig.data[clicked_trace_index].name

                # Update excluded isotopes based on the clicked trace
                if restyleData[0]['visible'][0] == 'legendonly' and clicked_trace_name not in self._excluded_isotopes:
                    self._excluded_isotopes.append(clicked_trace_name)
                elif restyleData[0]['visible'][0] == True and clicked_trace_name in self._excluded_isotopes:
                    self._excluded_isotopes.remove(clicked_trace_name)

                # Update DataFrame based on excluded isotopes
                updated_df = self.df.copy()
                updated_df = updated_df[~updated_df['Isotope'].isin(self._excluded_isotopes)]

                # Recalculate the regression and summary statistics
                self._add_regression_and_stats(updated_df)

                # Update trace visibility based on excluded isotopes
                for trace in self.fig.data:
                    if trace.name in self._excluded_isotopes:
                        trace.visible = 'legendonly'
                    else:
                        trace.visible = True

                return self.fig

            return dash.no_update

        @self._app.server.route('/shutdown', methods=['POST'])
        def shutdown():
            os.kill(os.getpid(), signal.SIGINT)  # Send the SIGINT signal to the current process
            return 'Server shutting down...'

    def show(self) -> None:
        """Display the interactive scatter plot with the interactive legend in a web browser."""
        port = _find_free_port()
        # Function to open the browser
        def open_browser():
            if not os.environ.get("WERKZEUG_RUN_MAIN"):
                print(f"Now running at http://localhost:{port}/")
                webbrowser.open(f"http://localhost:{port}/")

        # Silence the Flask development server logging
        log = open(os.devnull, 'w')
        # sys.stdout = log
        sys.stderr = log

        # Disable Flask development server warning
        os.environ['FLASK_ENV'] = 'development'

        # JavaScript code to detect when the tab or window is closed
        self._app.index_string = '''
        <!DOCTYPE html>
        <html>
            <head>
                {%metas%}
                <title>{%title%}</title>
                {%favicon%}
                {%css%}
            </head>
            <body style="margin: 0;">
                {%app_entry%}
                <footer>
                    {%config%}
                    {%scripts%}
                    <script type="text/javascript">
                        window.addEventListener("beforeunload", function (e) {
                            var xhr = new XMLHttpRequest();
                            xhr.open("POST", "/shutdown", false);
                            xhr.send();
                        });
                    </script>
                    {%renderer%}
                </footer>
            </body>
        </html>
        '''

        # Timer to open the browser shortly after the server starts
        if not config.generating_docs:
            threading.Timer(1, open_browser).start()
            self._app.run_server(debug=False, host='localhost', port=port)

    def save_state(self, filename: typing.Optional[Union[str, Path]]=None) -> Optional[dict]:
        """Save the current state of the interactive scatter plot to a file. This method saves the state of the interactive
        scatter plot as a dictionary containing the figure, DataFrame, excluded isotopes, index name, and nested status. This
        can be later reserialized by the :meth:`InteractiveScatterLegend.load_state` method, or the
        :func:`load_interactive_scatter_plot` function.
        
        Parameters
        ----------
        filename
            The filename to save the state to. If not provided, the state dictionary is returned instead of being saved to a 
            file.
            
        Returns
        -------
            * If ``filename`` is not provided, the state dictionary is returned. 
            * Otherwise, ``None`` is returned."""
        state = {
            'fig': self.fig.to_dict(),
            'df': self.df.to_dict(),
            'excluded_isotopes': self._excluded_isotopes,
            'index_name': self._index_name,
            'nested': self._interactive_scatter_plot._nested
        }
        if filename is None:
            return state
        else:
            with open(filename, 'wb') as f:
                pickle.dump(state, f)

    @classmethod
    def load_state(cls, filename: typing.Optional[Union[str, Path]]=None, data_dict: typing.Optional[dict]=None
                   ) -> InteractiveScatterLegend:
        """Load a saved state of an interactive scatter plot from a file. This method loads the state of the interactive scatter
        plot from a dictionary containing the figure, DataFrame, excluded isotopes, index name, and nested status. This can be
        reserialized by the :meth:`InteractiveScatterLegend.load_state` method, or the :func:`load_interactive_scatter_plot`
        function.

        Parameters
        ----------
        filename
            The filename of the saved state file to load the interactive scatter plot from.
        data_dict
            A dictionary containing the saved state of the interactive scatter plot. This is an alternative to providing the
            filename.
        
        Returns
        -------
            An instance of the interactive scatter plot that can be plotted with :meth:`InteractiveScatterLegend.show()`.
        """
        if filename is None and data_dict is None:
            raise ValueError("Either a filename or a data dictionary must be provided")
        if filename is not None:
            with open(filename, 'rb') as f:
                state = pickle.load(f)
        else:
            state = data_dict

        # Recreate the _InteractiveScatterPlotter instance from the saved state
        fig = go.Figure(state['fig'])
        index_name = state['index_name']
        nested = state['nested']
        interactive_scatter_plot = _InteractiveScatterPlotter(index_name, nested)
        interactive_scatter_plot.fig = fig

        # Recreate the InteractiveScatterLegend instance from the saved state
        instance = cls(interactive_scatter_plot, pd.DataFrame.from_dict(state['df']))
        instance._excluded_isotopes = state['excluded_isotopes']

        # Update trace visibility based on excluded isotopes
        for trace in instance.fig.data:
            if trace.name in instance._excluded_isotopes:
                trace.visible = 'legendonly'
            else:
                trace.visible = True

        return instance

    def write_html(self, filename: typing.Optional[Union[str, Path]]=None) -> Union[None, str]:
        """Save the current state of the interactive scatter plot to an HTML file. This method saves the current state of the
        interactive scatter plot to an HTML file that can be viewed in a web browser.

        Parameters
        ----------
        filename
            The filename to save the interactive scatter plot to.
            
        Notes
        -----
        Since the legend interactivity is implemented via python, saving the plot as HTML will only save the current state of
        the plot. The interactivity will not be preserved in the saved HTML file."""
        
        # Utilize Plotly's write_html to save the current state of the figure
        if filename is not None:
            self.fig.write_html(filename)
        else:
            return pio.to_html(self.fig, full_html=True)
        
    def to_image(self, filename: Union[str, Path]) -> None:
        """Write the initial state of the interactive plot to an image file. This function saves the plot as an 
        image by using selenium webdriver.
        
        Parameters
        ----------
        filename
            Name of the image file to save the matrix plot to. The file extension should be ``'.png'``, ``.jpg``, etc."""
        with tempfile.NamedTemporaryFile(delete=False) as f:
            html_filename = f.name + '.html'
        self.write_html(html_filename)
        _capture_html_as_image(html_filename, filename)
        os.remove(html_filename)



"""Various utility functions for creating interactive plots, including functions for generating plot objects from
contributions and perturbations. These functions are used in the interactive plotting functions in the ``viz`` module."""

import socket
from tsunami_ip_utils._utils import _filter_redundant_reactions, _isotope_reaction_list_to_nested_dict
import numpy as np
from typing import Dict, List, Tuple
from uncertainties import unumpy
from pathlib import Path
from uncertainties import ufloat
import tsunami_ip_utils
from selenium import webdriver
from selenium.webdriver.chrome.service import Service
from webdriver_manager.chrome import ChromeDriverManager
import time
from selenium.webdriver.common.by import By

def _find_free_port() -> int:
    """Finds a free port on localhost for running a Flask/Dash server. This is done by creating a socket and binding it
    to an available port. The socket is then closed and the port number is returned.
    
    Returns
    -------
        A free port number."""
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.bind(("", 0))  # Let the OS pick an available port
    port = s.getsockname()[1]
    s.close()
    return port

def _determine_plot_type(contributions: List[ dict ], plot_redundant_reactions: bool) -> Tuple[dict, bool]:
    """Determines whether the contributions are nuclide-wise or nuclide-reaction-wise and whether to plot redundant
    reactions or not. Then converts the contributions (list of dictionaries) to a dictionary of contributions keyed by
    isotope then by reaction type if necessary.
    
    Parameters
    ----------
    contributions
        List of dictionaries containing the contributions to the similarity parameter for each
        nuclide or nuclide-reaction pair.
    plot_redundant_reactions
        Whether to plot redundant reactions (or irrelevant reactions) when considering
        nuclide-reaction-wise contributions.
        
    Returns
    -------
        - contributions
            Contributions to the similarity parameter keyed by isotope then by reaction type (if necessary).
        - nested_plot
            Whether the plot is nested by nuclide then by reaction type."""
    if 'reaction_type' in contributions[0]: # Nuclide-reaction-wise contributions
        nested_plot = True # Nested plot by nuclide then by reaction type

        # Create a dictionary of contributions keyed by isotope then by reaction type
        contributions = _isotope_reaction_list_to_nested_dict(contributions, 'contribution')

        # If viewing nuclide-reaction wise contributions, it's important (at least for the visualizations in this function)
        # that if viewing the true contributions to the nuclide total, that redundant interactions (e.g. capture and fission
        # + (n, g)) and irrelevant interactions (e.g. chi and nubar) are not plotted.

        if not plot_redundant_reactions:
            # Remove redundant interactions
            contributions = _filter_redundant_reactions(contributions)
    else: # Nuclide-wise contributions
        nested_plot = False
        contributions = { contribution['isotope']: contribution['contribution'] for contribution in contributions }

    return contributions, nested_plot

def generate_plot_objects_array_from_contributions(contributions: Dict[ str, List[ unumpy.uarray ] ], integral_index_name: str, 
                                                   **kwargs: dict) -> np.ndarray:
    """Generate a matrix of plot objects (for creating a matrix plot) for the given contributions to an arbitrary integral index.
    This is valid for plots of :math:`\\Delta k/k` contributions, :math:`E` contributions, :math:`c_k` contributions, etc..
    
    Parameters
    ----------
    contributions
        Dictionary of a list of contributions to the integral index keyed by application or experiment.
    integral_index_name
        Name of the integral index being plotted.
    kwargs
        Additional keyword arguments. The following are supported:
        
        - diagonal_type (str)
            Type of plot to create on the diagonal. Default is ``'interactive_pie'`` which creates an interactive
            pie chart.
        - off_diagonal_type (str)
            Type of plot to create off the diagonal. Default is ``'interactive_scatter'`` which creates an interactive
            scatter plot.
        - interactive_contribution_legend (bool)
            Whether to make the legend interactive for the contribution plots. Default is ``True``.
        - interactive_correlation_legend (bool)
            Whether to make the legend interactive for the correlation plots. Default is ``True``.
            
    Returns
    -------
        2D numpy array of plot objects to be plotted with the :func:`tsunami_ip_utils.viz.viz.matrix_plot` function."""
    from tsunami_ip_utils.viz import contribution_plot, correlation_plot # Import here to avoid circular import
    
    # Get options for legend interactivity and the diagonal plot type if supplied
    diagonal_type = kwargs.get('diagonal_type', 'interactive_pie')
    off_diagonal_type = kwargs.get('off_diagonal_type', 'interactive_scatter')
    interactive_correlation_legend = kwargs.get('interactive_correlation_legend', True)
    interactive_contribution_legend = kwargs.get('interactive_contribution_legend', True)
    
    num_applications = len(contributions['application'])
    num_experiments = len(contributions['experiment'])

    # Construct plot matrix
    plot_objects_array = np.empty( ( num_applications, num_experiments ), dtype=object )

    for application_index in range(num_applications):
        for experiment_index in range(num_experiments):
            if experiment_index == application_index:
                # On the diagonal, make a contribution plot, as a correlation plot is not useful when comparing the same
                # application and experiment
                plot_objects_array[application_index, experiment_index] = \
                contribution_plot(
                    contributions['application'][application_index],
                    plot_type=diagonal_type,
                    integral_index_name=integral_index_name,
                    interactive_legend=interactive_contribution_legend,     
                )
            else:
                plot_objects_array[application_index, experiment_index] =  \
                correlation_plot(
                    contributions['application'][application_index], 
                    contributions['experiment'][experiment_index], 
                    plot_type=off_diagonal_type,
                    integral_index_name=integral_index_name, 
                    plot_redundant_reactions=True, 
                    interactive_legend=interactive_correlation_legend
                )

    return plot_objects_array

def generate_plot_objects_array_from_perturbations(points_array: np.ndarray) -> np.ndarray:
    """Generate a matrix of plot objects (for creating a matrix plot) from a numpy array of perturbation points. This is
    used for a matrix of perturbation plots only.
    
    Parameters
    ----------
    points_array
        2D numpy array of points generated from the perturbation test. Shape ``(num_applications, num_experiments)``.
            
    Returns
    -------
        2D numpy array of plot objects to be plotted with the :func:`tsunami_ip_utils.viz.viz.matrix_plot` function."""
    from tsunami_ip_utils.viz import perturbation_plot # Import here to avoid circular import
    
    # Construct plot matrix
    num_applications, num_experiments, _, _ = np.shape(points_array)
    plot_objects_array = np.empty( ( num_applications, num_experiments ), dtype=object)

    for i, row in enumerate(points_array):
        for j, _ in enumerate(row):
            fig = perturbation_plot(points_array[i, j])
            plot_objects_array[i, j] = fig

    return plot_objects_array

def _capture_html_as_image(html_file: Path, output_image: Path, matrix: bool=False) -> None:
    """Captures an HTML file as an image using Selenium WebDriver. This is useful for saving interactive plots as static
    images.

    Parameters
    ----------
    html_file
        Path to the HTML file to capture.
    output_image
        Path to save the output image
    matrix
        Whether or not the html file is a matrix plot. If ``True``, the function searches for a div element with a specific
        id to determine the necessary size of the webbrowser.
    """
    # Setup WebDriver
    service = Service(ChromeDriverManager().install())
    options = webdriver.ChromeOptions()
    options.add_argument("--headless")
    driver = webdriver.Chrome(service=service, options=options)

    # Load HTML file
    html_file = Path("file://" + str(html_file))
    driver.get(str(html_file))

    def get_scroll_dimension(axis):
        return driver.execute_script(f"return document.body.parentNode.scroll{axis}")
    
    def get_element_scroll_dimension(element_id, property):
        script = f"return document.getElementById('{element_id}').scroll{property};"
        return driver.execute_script(script)

    # get the page scroll dimensions
    if matrix:
        width = get_element_scroll_dimension("matrix-plot", "Width")
    else:
        width = get_scroll_dimension("Width")
    height = get_scroll_dimension("Height")

    # set the browser window size
    driver.set_window_size(width, height)
    time.sleep(1)

    # Take screenshot
    driver.save_screenshot(str(output_image))
    driver.quit()


"""This module contains the :class:`InteractiveMatrixPlot` class, which is a wrapper around a Dash app that displays a matrix of
interactive plots. The matrix is constructed from a 2D numpy array of plot objects, where each plot object is an instance of
either :class:`tsunami_ip_utils.viz.scatter_plot.InteractiveScatterLegend` or 
:class:`tsunami_ip_utils.viz.pie_plot.InteractivePieLegend`. The matrix is displayed in a Dash app, where each plot object is
displayed in a separate cell. The user can interact with the plots by clicking on the legend to hide or show traces. The user
can also save the state of the interactive matrix plot to a pickle file and load it back later. The :func:`load_interactive_matrix_plot`
function is a convenience function that loads an interactive matrix plot from a saved state pickle file."""

from __future__ import annotations
import dash
from dash import dcc, html
from dash.dependencies import Input, Output, State
import numpy as np
from pathlib import Path
import plotly.graph_objects as go
from .scatter_plot import InteractiveScatterLegend
from .pie_plot import InteractivePieLegend
import webbrowser
import os
import sys
import threading
from .plot_utils import _find_free_port
import pickle
import tsunami_ip_utils
from typing import Union, List, Dict, Optional
import tsunami_ip_utils.config as config
import time, requests
import signal
from selenium import webdriver
from selenium.webdriver.chrome.options import Options
import multiprocessing
import tempfile
from tsunami_ip_utils.viz.plot_utils import _capture_html_as_image
import base64
import io
import matplotlib
import matplotlib.pyplot as plt
import plotly.io as pio

# Style constants
GRAPH_STYLE = {
    'flex': '1',
    'minWidth': '800px',
    'height': '500px',
    'padding': '10px',
    'borderRight': '1px solid black',
    'borderBottom': '1px solid black',
    'borderTop': '0px',
    'borderLeft': '0px'
}

def _create_app(external_stylesheets):
    app = dash.Dash(__name__, external_stylesheets=external_stylesheets)
    return app

def _create_column_headers(num_cols: int, labels: Optional[List[str]]=None) -> None:
    """Create column headers for the matrix plot. Each column header is a div element with the text 'Application i' where i is
    the column index. The column headers are styled to be centered and have a border on the right and bottom.
    
    Parameters
    ----------
    num_cols
        Number of columns in the matrix plot.
    labels
        List of labels for the columns of the matrix plot. Default is ``None``."""
    if labels is None:
        div = [html.Div(
            f'Application {i+1}', 
            style={
                'flex': '1', 
                'minWidth': '800px', 
                'textAlign': 'center', 
                'padding': '10px', 
                'borderRight': '1px solid black', 
                'borderBottom': '1px solid black', 
                'display': 'flex', 
                'alignItems': 'center', 
                'justifyContent': 'center'
            }
        ) for i in range(num_cols)]
    else:
        div = [html.Div(
            label, 
            style={
                'flex': '1', 
                'minWidth': '800px', 
                'textAlign': 'center', 
                'padding': '10px', 
                'borderRight': '1px solid black', 
                'borderBottom': '1px solid black', 
                'display': 'flex', 
                'alignItems': 'center', 
                'justifyContent': 'center'
            }
        ) for label in labels]
    return div

def _create_row_label(i: int, label: Optional[str]=None) -> html.Div:
    """Create a row label for the matrix plot. The row label is a div element with the text 'Experiment i' where i is the row
    index. The row label is styled to be centered and have a border on the right and bottom. The text is rotated -90 degrees to
    make it vertical.
    
    Parameters
    ----------
    i
        Row index of the matrix plot.
    label
        An optional label for the row (to use instead of ``Experiment i``). Default is ``None``.
        
    Returns
    -------
        A div element representing the row label."""
    if label is None:
        div = html.Div(
            html.Span(
                f'Experiment {i+1}',
                style={
                    'display': 'block',
                    'overflow': 'visible',
                    'transform': 'rotate(-90deg)',
                    'transformOrigin': 'center',
                    'whiteSpace': 'nowrap',
                }
            ), 
            style={
                'flex': 'none',
                'width': '50px', 
                'textAlign': 'center', 
                'marginRight': '0', 
                'padding': '10px', 
                'borderRight': '1px solid black', 
                'borderBottom': '1px solid black', 
                'display': 'flex', 
                'alignItems': 'center', 
                'justifyContent': 'center'
            }
        )
    else:
        div = html.Div(
            html.Span(
                label,
                style={
                    'display': 'block',
                    'overflow': 'visible',
                    'transform': 'rotate(-90deg)',
                    'transformOrigin': 'center',
                    'whiteSpace': 'nowrap',
                }
            ), 
            style={
                'flex': 'none',
                'width': '50px', 
                'textAlign': 'center', 
                'marginRight': '0', 
                'padding': '10px', 
                'borderRight': '1px solid black', 
                'borderBottom': '1px solid black', 
                'display': 'flex', 
                'alignItems': 'center', 
                'justifyContent': 'center'
            }
        )

    return div

def _create_plot_element(i: int, j: int, plot_object: Union[InteractiveScatterLegend, InteractivePieLegend, go.Figure]
                        ) -> Union[dcc.Graph, html.Iframe]:
    """Create a plot element based on the plot object. If the plot object is an instance of 
    :class:`tsunami_ip_utils.viz.scatter_plot.InteractiveScatterLegend`, or a ``plotly.graph_objects.Figure``, 
    the plot element is a ``dcc.Graph`` element. If the plot object is an instance of
    :class:`tsunami_ip_utils.viz.pie_plot.InteractivePieLegend`, the plot element is an ``html.Iframe`` element.
    
    """
    if isinstance(plot_object, InteractiveScatterLegend):
        graph_id = f"interactive-scatter-{i}-{j}"
        return dcc.Graph(id=graph_id, figure=plot_object.fig, style=GRAPH_STYLE)
    elif isinstance(plot_object, InteractivePieLegend):
        with plot_object._app.test_client() as client:
            response = client.get('/')
            html_content = response.data.decode('utf-8')
            return html.Iframe(srcDoc=html_content, style=GRAPH_STYLE)
    elif isinstance(plot_object, tuple):
        if isinstance(plot_object[0], matplotlib.figure.Figure):
            # Convert matplotlib figure to PNG image
            buf = io.BytesIO()
            plot_object[0].tight_layout()
            plot_object[0].savefig(buf, format='png')
            buf.seek(0)
            plt.close(plot_object[0])  # Close the figure to prevent it from being captured
            encoded_image = base64.b64encode(buf.read()).decode("utf-8")
            return html.Img(src='data:image/png;base64,{}'.format(encoded_image), style=GRAPH_STYLE)
    else:
        return dcc.Graph(figure=plot_object, style=GRAPH_STYLE)

def _create_update_figure_callback(app, graph_id, app_instance):
    @app.callback(
        Output(graph_id, 'figure'),
        Input(graph_id, 'restyleData'),
        State(graph_id, 'figure')
    )
    def update_figure_on_legend_click(restyleData, current_figure_state):
        if restyleData and 'visible' in restyleData[0]:
            current_fig = go.Figure(current_figure_state)

            # Get the index of the clicked trace
            clicked_trace_index = restyleData[1][0]

            # Get the name of the clicked trace
            clicked_trace_name = current_fig.data[clicked_trace_index].name

            # Update excluded isotopes based on the clicked trace
            if restyleData[0]['visible'][0] == 'legendonly' and clicked_trace_name not in app_instance._excluded_isotopes:
                app_instance._excluded_isotopes.append(clicked_trace_name)
            elif restyleData[0]['visible'][0] == True and clicked_trace_name in app_instance._excluded_isotopes:
                app_instance._excluded_isotopes.remove(clicked_trace_name)

            # Update DataFrame based on excluded isotopes
            updated_df = app_instance.df.copy()
            updated_df = updated_df[~updated_df['Isotope'].isin(app_instance._excluded_isotopes)]

            # Recalculate the regression and summary statistics
            app_instance._add_regression_and_stats(updated_df)

            # Update trace visibility based on excluded isotopes
            for trace in app_instance.fig.data:
                if trace.name in app_instance._excluded_isotopes:
                    trace.visible = 'legendonly'
                else:
                    trace.visible = True

            return app_instance.fig

        return dash.no_update

def _generate_layout(app: dash.Dash, rows: List[html.Div]) -> None:
    """Generate the layout of the Dash app. The layout consists of an H1 element with the title 'Matrix of Plots', followed by
    a div element containing the rows of the matrix plot. The rows are displayed in a flex column with horizontal scrolling.
    The layout also includes a JavaScript script that listens for window resize events and resizes the Plotly plots
    accordingly.
    
    Parameters
    ----------
    app
        Dash app object.
    rows
        List of div elements representing the rows of the matrix plot."""
    app.layout = html.Div([
        html.H1("Matrix of Plots", style={'textAlign': 'center', 'marginLeft': '121px'}),
        html.Div(rows, id='matrix-plot',style={'display': 'flex', 'flexDirection': 'column', 'width': '100%', 'overflowX': 'auto'}),
        html.Script("""
        window.addEventListener('resize', function() {
            const graphs = Array.from(document.querySelectorAll('.js-plotly-plot'));
            graphs.forEach(graph => {
                Plotly.Plots.resize(graph);
            });
        });
        """)
    ])

class InteractiveMatrixPlot:
    """Interactive matrix plot class that displays a matrix of interactive plots in a Dash app. The matrix is constructed from
    a 2D numpy array of plot objects, where each plot object is an instance of either
    :class:`tsunami_ip_utils.viz.scatter_plot.InteractiveScatterLegend` or
    :class:`tsunami_ip_utils.viz.pie_plot.InteractivePieLegend`. The matrix is displayed in a Dash app, where each plot object is
    displayed in a separate cell. The user can interact with the plots by clicking on the legend to hide or show traces. The user
    can also save the state of the interactive matrix plot to a pickle file and load it back later."""
    _app: dash.Dash
    """Dash app object that displays the interactive matrix plot."""
    _plot_objects_array: np.ndarray
    """2D numpy array of plot objects to be displayed in the matrix plot."""
    def __init__(self, app: dash.Dash, plot_objects_array: np.ndarray, labels: Optional[Dict[str, List]]=None) -> None:
        """Initialize the InteractiveMatrixPlot object with the Dash app and the 2D numpy array of plot objects.
        
        Parameters
        ----------
        app
            Dash app object that displays the interactive matrix plot.
        plot_objects_array
            2D numpy array of plot objects to be displayed in the matrix plot.
        labels
            Dictionary of lists containing the labels for the rows and columns of the matrix plot. Keys are ``'applications'`` 
            and ``'experiments'``. Default is ``None``.
        """
        self._app = app
        self._plot_objects_array = plot_objects_array
        self._labels = labels
    
    def _open_browser(self, port: int) -> None:
        """Open the browser to display the interactive matrix plot.
        
        Parameters
        ----------
        port
            Port number of the Flask server."""
        print(f"Now running at http://localhost:{port}/")
        webbrowser.open(f"http://localhost:{port}/")
        pass

    def show(self, open_browser: bool=True, silent: bool=False):
        """Start the Flask server and open the browser to display the interactive sunburst chart
        
        Parameters
        ----------
        open_browser
            Whether to open the browser automatically to display the chart.
        silent
            Whether to suppress Flask's startup and runtime messages."""
        # Suppress Flask's startup and runtime messages by redirecting them to dev null
        log = open(os.devnull, 'w')
        sys.stderr = log
        if silent:
            sys.stdout = log

        port = _find_free_port()
        if not config.generating_docs:
            if open_browser:
                threading.Timer(1, self._open_browser(port)).start()
            self._app.run(host='localhost', port=port, debug=False)
    
    def save_state(self, filename: Union[str, Path]) -> None:
        """Save the state of the interactive matrix plot to a pickle file. The state includes the 2D numpy array of plot objects
        and the types of the plot objects.
        
        Parameters
        ----------
        filename
            Name of the pickle file to save the state to. The file extension should be ``'.pkl'``."""
        # Serialize interactive plots in the plot objects array
        self.plot_types = np.empty_like(self._plot_objects_array, dtype=object)
        for i, row in enumerate(self._plot_objects_array):
            for j, plot_object in enumerate(row):
                if isinstance(plot_object, InteractiveScatterLegend):
                    self._plot_objects_array[i,j] = plot_object.save_state()
                    self.plot_types[i,j] = "InteractiveScatterLegend"
                elif isinstance(plot_object, InteractivePieLegend):
                    self._plot_objects_array[i,j] = plot_object.save_state()
                    self.plot_types[i,j] = "InteractivePieLegend"

        with open(filename, 'wb') as f:
            pickle.dump( ( self._plot_objects_array, self.plot_types ) , f)

    @classmethod
    def load_state(self, filename: Union[str, Path]) -> InteractiveMatrixPlot:
        """Loads an interactive matrix plot from a saved state pickle file.
        
        Parameters
        ----------
        filename
            Name of the pickle file to load the state from.
            
        Returns
        -------
            An reserialized instance of the :class:`InteractiveMatrixPlot` class."""
        with open(filename, 'rb') as f:
            plot_objects_array, plot_types = pickle.load(f)
            # Reserialize interactive scatter legends
            for i, row in enumerate(plot_objects_array):
                for j, plot_object in enumerate(row):
                    if plot_types[i,j] == "InteractiveScatterLegend":
                        plot_objects_array[i,j] = InteractiveScatterLegend.load_state(data_dict=plot_object)
                    elif plot_types[i,j] == "InteractivePieLegend":
                        plot_objects_array[i,j] = InteractivePieLegend.load_state(data_dict=plot_object)

        return _interactive_matrix_plot(plot_objects_array)

    def write_html(self, filename: Optional[Union[str, Path]]=None) -> None:
        """Write the HTML content of the interactive matrix plot to a file.

        Parameters
        ----------
        filename
            Name of the HTML file to write the content to. The file extension should be `'.html'`.
        """
        # Create a new Dash app instance
        app = _create_app(external_stylesheets=[])

        num_rows = self._plot_objects_array.shape[0]
        num_cols = self._plot_objects_array.shape[1]

        column_headers = _create_column_headers(num_cols, self._labels['applications']) if self._labels else _create_column_headers(num_cols)
        header_row = html.Div([html.Div('', style={'flex': 'none', 'width': '71px', 'borderBottom': '1px solid black'})] + column_headers, style={'display': 'flex'})

        rows = [header_row]
        for i in range(num_rows):
            row = [_create_row_label(i, self._labels['experiments'][i])] if self._labels else [_create_row_label(i)]
            for j in range(num_cols):
                plot_object = self._plot_objects_array[i, j]
                if isinstance(plot_object, InteractiveScatterLegend):
                    plot_element = html.Iframe(srcDoc=plot_object.write_html(), style=GRAPH_STYLE)
                elif isinstance(plot_object, InteractivePieLegend):
                    plot_element = html.Iframe(srcDoc=plot_object.write_html(), style=GRAPH_STYLE)
                elif isinstance(plot_object, tuple):
                    if isinstance(plot_object[0], matplotlib.figure.Figure):
                        # Convert matplotlib figure to PNG image
                        buf = io.BytesIO()
                        plot_object[0].tight_layout()
                        plot_object[0].savefig(buf, format='png')
                        buf.seek(0)
                        plt.close(plot_object[0])  # Close the figure to prevent it from being captured
                        encoded_image = base64.b64encode(buf.read()).decode("utf-8")
                        plot_element = html.Img(src='data:image/png;base64,{}'.format(encoded_image), style=GRAPH_STYLE)
                else:
                    html_string = pio.to_html(plot_object, full_html=False)
                    plot_element = html.Iframe(srcDoc=html_string, style=GRAPH_STYLE)
                row.append(plot_element)
            rows.append(html.Div(row, style={'display': 'flex'}))

        _generate_layout(app, rows)

        port = _find_free_port()
        def run_app():
            log = open(os.devnull, 'w')
            sys.stdout = log
            sys.stderr = log
            app.run_server('localhost', port, debug=False)

        # Start the app in a separate thread
        process = multiprocessing.Process(target=run_app)
        process.start()

        # Set up a headless Chrome browser using Selenium
        chrome_options = Options()
        chrome_options.add_argument("--headless")  # Run Chrome in headless mode
        driver = webdriver.Chrome(options=chrome_options)

        try:
            # Load the Dash app in the headless browser
            driver.get(f"http://localhost:{port}")

            # Wait for the app to finish loading (you can adjust the sleep time if needed)
            time.sleep(1)

            # Capture the fully rendered HTML content
            html_content = driver.page_source

            # Write the HTML content to the specified file
            if filename is not None:
                with open(filename, "w") as file:
                    file.write(html_content)
            else:
                return html_content

        finally:
            # Quit the headless browser
            driver.quit()

            # Stop the Dash server
            requests.post(f"http://localhost:{port}/shutdown")
            process.terminate()
            process.join()
    
    def to_image(self, filename: Union[str, Path]):
        """Write the initial state of the InteractiveMatrixPlot to an image file. This function saves the matrix plot as an 
        image by using selenium webdriver.
        
        Parameters
        ----------
        filename
            Name of the image file to save the matrix plot to. The file extension should be ``'.png'``, ``.jpg``, etc."""
        with tempfile.NamedTemporaryFile(delete=False) as f:
            html_filename = f.name + '.html'
        self.write_html(html_filename)
        _capture_html_as_image(html_filename, filename, matrix=True)
        os.remove(html_filename)


def load_interactive_matrix_plot(filename):
    """Loads an interactive matrix plot from a saved state pickle file. This function is purely for convenience and is a
    wrapper of the :meth:`InteractiveMatrixPlot.load_state` class method"""
    return InteractiveMatrixPlot.load_state(filename)


def _interactive_matrix_plot(plot_objects_array: np.ndarray, labels: Optional[Dict[str, List]]) -> InteractiveMatrixPlot:
    """Create an interactive matrix plot from a 2D numpy array of plot objects. This function creates a Dash app that displays
    the matrix plot. The matrix is constructed from the plot objects array, where each plot object is an instance of either
    :class:`tsunami_ip_utils.viz.scatter_plot.InteractiveScatterLegend` or :class:`tsunami_ip_utils.viz.pie_plot.InteractivePieLegend`.
    The matrix is displayed in a Dash app, where each plot object is displayed in a separate cell. The user can interact with
    the plots by clicking on the legend to hide or show traces.

    Parameters
    ----------
    plot_objects_array
        2D numpy array of plot objects to be displayed in the matrix plot.
    labels
        Dictionary of lists containing the labels for the rows and columns of the matrix plot. Keys are ``'applications'`` and 
        ``'experiments'``. Default is ``None``. Most commonly, this is the sdf filename for the given application/experiment.

    Returns
    -------
        An instance of the :class:`InteractiveMatrixPlot` class that wraps the Dash app displaying the matrix plot.
    """
    current_directory = Path(__file__).parent
    external_stylesheets = [str(current_directory / 'css' / 'matrix_plot.css')]
    app = _create_app(external_stylesheets)
    @app.server.route('/shutdown', methods=['POST'])
    def shutdown():
        """Function to shutdown the server"""
        os.kill(os.getpid(), signal.SIGINT)  # Send the SIGINT signal to the current process
        return 'Server shutting down...'

    num_rows = plot_objects_array.shape[0]
    num_cols = plot_objects_array.shape[1]

    column_headers = _create_column_headers(num_cols, labels['applications']) if labels else _create_column_headers(num_cols)
    header_row = html.Div([html.Div('', style={'flex': 'none', 'width': '71px', 'borderBottom': '1px solid black'})] + column_headers, style={'display': 'flex'})

    rows = [header_row]
    for i in range(num_rows):
        if labels is not None:
            row = [_create_row_label(i, labels['experiments'][i])]
        else:
            row = [_create_row_label(i)]
        for j in range(num_cols):
            plot_object = plot_objects_array[i, j]
            plot_element = _create_plot_element(i, j, plot_object) if plot_object else html.Div('Plot not available', style=GRAPH_STYLE)
            row.append(plot_element)
            if isinstance(plot_object, InteractiveScatterLegend):
                _create_update_figure_callback(app, f"interactive-scatter-{i}-{j}", plot_object)
        rows.append(html.Div(row, style={'display': 'flex'}))

    _generate_layout(app, rows)
    return InteractiveMatrixPlot(app, plot_objects_array, labels=labels)


from tsunami_ip_utils.viz.viz import correlation_plot, contribution_plot, perturbation_plot, matrix_plot
from tsunami_ip_utils.viz.matrix_plot import load_interactive_matrix_plot
from tsunami_ip_utils.viz.scatter_plot import load_interactive_scatter_plot
from tsunami_ip_utils.viz.viz import generate_heatmap_from_comparison


from abc import ABC, abstractmethod
from typing import Any

class _Plotter(ABC):
    @abstractmethod
    def _create_plot(self, data, nested) -> None:
        pass

    @abstractmethod
    def _add_to_subplot(self, fig: Any, position: tuple) -> Any:
        """Add the plot (corresponding to the class instance) to the given figure subplot at the given position.
        
        Parameters
        ----------
        fig
            The figure to add the plot to.
        position
            The position of the subplot to add the plot to.
            
        Returns
        -------
            The figure with the plot added."""
        pass

    @abstractmethod
    def _get_plot(self) -> Any:
        """Return the 'plot' object.
        
        Returns
        -------
            The plot object."""
        pass

    @abstractmethod
    def _style(self) -> None:
        """Style the plot."""
        pass


"""Functionality related to matplotlib bar plots of contributions to integral indices."""

import matplotlib.pyplot as plt
from matplotlib.figure import Figure
from matplotlib.axes import Axes
from ._base_plotter import _Plotter
import numpy as np
from typing import Dict, Union, Tuple
from uncertainties import ufloat
from tsunami_ip_utils._utils import _isotope_reaction_list_to_nested_dict
import matplotlib as mpl

mpl.rcParams['hatch.linewidth'] = 0.25

class _BarPlotter(_Plotter):
    """Class for creating bar plots of contributions to integral indices on a nuclide-wise and nuclide-reaction-wise basis."""
    _fig: Figure
    """The figure object for the plot."""
    _axs: plt.Axes
    """The axes object for the plot."""
    _index_name: str
    """The name of the integral index being plotted."""
    def __init__(self, integral_index_name: str, plot_redundant: bool=False, **kwargs):
        """Initializes a bar plot of the contributions to the given integral index.
        
        Parameters
        ----------
        integral_index_name
            The name of the integral index whose contributions are to be plotted.
        plot_redundant
            Wether to include redundant/irrelevant reactions in the plot. NOTE: this only applies to nested plots, and
            only affects the plot title; it is expected that the provided data is consistent with the flag.
            
        Notes
        -----
        * Redundant reactions are defined as those which are derived from other reactions, e.g. 'total' and 'capture' reactions
          in SCALE.
        * Irrelevant reactions are defined as those which are not directly cross sections (but rather other nuclear data 
          parameters), e.g. 'chi' in SCALE.
        * A flag for including/excluding redundant/irrelevant reactions was provided since, if the user is expecting the
          the contributions to add up nicely, then the redundant reactions should be excluded, and if only cross sections are
          being considered, then the irrelevant reactions should be excluded.
        """
        self._index_name = integral_index_name
        self.plot_redundant = plot_redundant

    def _create_plot(self, contributions: Union[Dict[str, ufloat], Dict[str, Dict[str, ufloat]]], nested: bool):
        """Creates a bar plot of the given contributions to the integral index.
        
        Parameters
        ----------
        contributions
            * If ``nested`` is ``False``, then this should be a dictionary of the form ``{nuclide: contribution}``, where 
              contribution is a ``ufloat`` object representing the contribution of the nuclide to the integral index.
            * If ``nested`` is ``True``, then this should be a dictionary of the form ``{nuclide: {reaction: contribution}}``,
              where contribution is a ``ufloat`` object representing the contribution of the nuclide to the integral index 
              through the given reaction.
        nested
            Wether the contributions are on a reaction-wise basis or not."""
        self.nested = nested
        self._fig, self._axs = plt.subplots()
        if nested:
            self._nested_barchart(contributions)
        else:
            self._barchart(contributions)

        self._style()

    def _get_plot(self) -> Tuple[Figure, Axes]:
        return self._fig, self._axs
        
    def _add_to_subplot(self, fig, position) -> Figure:
        return fig.add_subplot(position, sharex=self._axs, sharey=self._axs)
        
    def _barchart(self, contributions: Dict[str, ufloat]):
        """Create a bar chart of the contributions to the integral index on a nuclide-wise basis.
        
        Parameters
        ----------
        contributions
            A dictionary of the form ``{nuclide: contribution}``, where contribution is a ``ufloat`` object representing the
            contribution of the nuclide to the integral index."""
        sorted_contributions = sorted(contributions.items(), key=lambda item: abs(item[1].n), reverse=True)
        labels = [item[0] for item in sorted_contributions]
        values = [item[1].n for item in sorted_contributions]
        errors = [item[1].s for item in sorted_contributions]

        self._axs.bar(labels, values, yerr=errors, capsize=5, error_kw={'elinewidth': 0.5})
        self._axs.axhline(0, color='black', linewidth=0.8)  # Add a line at y=0 for clarity

    def _nested_barchart(self, contributions: Dict[str, Dict[str, ufloat]]):
        """Create a bar chart of the contributions to the integral index on a nuclide-reaction-wise basis."""
        # Collect all unique reactions across all nuclides
        all_reactions = set()
        for nuclide_reactions in contributions.values():
            all_reactions.update(nuclide_reactions.keys())
        all_reactions = sorted(all_reactions)

        # Colors for each reaction type
        cmap = plt.get_cmap('rainbow')
        colors = cmap(np.linspace(0, 1, len(all_reactions)))

        # Calculate total values for each nuclide and sort the data
        total_values = {label: sum(contributions[label][r].n for r in contributions[label]) for label in contributions}
        sorted_data = sorted(total_values.items(), key=lambda x: abs(x[1]), reverse=True)
        sorted_labels = [item[0] for item in sorted_data]
        sorted_total_values = [item[1] for item in sorted_data]

        # Variables to hold the bar positions and labels
        indices = range(len(sorted_labels))
        labels = sorted_labels

        # Bottom offset for each stack
        bottoms_pos = np.zeros(len(sorted_labels))
        bottoms_neg = np.zeros(len(sorted_labels))

        for reaction, color in zip(all_reactions, colors):
            values = [(contributions[nuclide].get(reaction, ufloat(0, 0)).n if reaction in contributions[nuclide] else 0) for nuclide in sorted_labels]
            errs = [(contributions[nuclide].get(reaction, ufloat(0, 0)).s if reaction in contributions[nuclide] else 0) for nuclide in sorted_labels]
            pos_values = [max(0, v) for v in values]
            neg_values = [min(0, v) for v in values]
            self._axs.bar(indices, pos_values, label=reaction, bottom=bottoms_pos, color=color,
                        yerr=errs, capsize=5, error_kw={'capthick': 0.5})
            self._axs.bar(indices, neg_values, bottom=bottoms_neg, color=color,
                        yerr=errs, capsize=5, error_kw={'capthick': 0.5})
            bottoms_pos += pos_values
            bottoms_neg += neg_values

        # Adding 'effective' box with dashed border
        for idx, val in zip(indices, sorted_total_values):
            self._axs.bar(idx, abs(val), bottom=0 if val > 0 else val, color='none', edgecolor='black', hatch='///', linewidth=0.25)

        self._axs.set_xticks(indices)
        self._axs.set_xticklabels(labels)
        self._axs.legend(loc='upper left', bbox_to_anchor=(1, 1), fontsize='small')

    def _style(self):
        if self.plot_redundant and self.nested:
            title_text = f'Contributions to {self._index_name} (including redundant/irrelvant reactions)'
        else:
            title_text = f'Contributions to {self._index_name}'
        self._axs.set_ylabel(f"Contribution to {self._index_name}")
        self._axs.set_xlabel("Isotope")
        self._axs.grid(True, which='both', axis='y', color='gray', linestyle='-', linewidth=0.5)
        self._axs.set_title(title_text)
        ticklabels = self._axs.get_xticklabels()
        self._axs.set_xticks(range(len(ticklabels))) # Have to reset these otherwise the next line gives a warning
        self._axs.set_xticklabels(ticklabels, rotation=45, rotation_mode='anchor', ha='right', va='top') 


"""Tools for creating pie charts of contributions to integral indices."""

from __future__ import annotations
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import plotly.express as px
from ._base_plotter import _Plotter
from plotly.subplots import make_subplots
import os, sys, signal
import threading
import webbrowser
from flask import Flask, render_template_string
import uuid
from .plot_utils import _find_free_port
import pickle
from typing import Tuple, Dict, Union, Optional
from uncertainties import ufloat
from matplotlib.figure import Figure
from matplotlib.axes import Axes
import plotly
from pathlib import Path
from tsunami_ip_utils import config
import tempfile
from tsunami_ip_utils.viz.plot_utils import _capture_html_as_image

plt.rcParams['hatch.linewidth'] = 0.6

class _PiePlotter(_Plotter):
    """A class for creating static pie charts of contributions to integral indices."""
    _index_name: str
    """The name of the integral index whose contributions are being plotted (e.g. ``'E'`` or ``'c_k'``)."""
    _plot_redundant: bool
    """Whether to plot redundant/irrelevant reactions in the pie chart."""
    def __init__(self, integral_index_name, plot_redudant=False, **kwargs):
        """Initializes a pie chart of the contributions to the given integral index.
        
        Parameters
        ----------
        integral_index_name
            The name of the integral index whose contributions are to be plotted.
        plot_redundant
            Wether to include redundant/irrelevant reactions in the plot. NOTE: this only applies to nested plots, and
            only affects the plot title; it is expected that the provided data is consistent with the flag.
            
        Notes
        -----
        * Redundant reactions are defined as those which are derived from other reactions, e.g. 'total' and 'capture' reactions
          in SCALE.
        * Irrelevant reactions are defined as those which are not directly cross sections (but rather other nuclear data 
          parameters), e.g. 'chi' in SCALE.
        * A flag for including/excluding redundant/irrelevant reactions was provided since, if the user is expecting the
          the contributions to add up nicely, then the redundant reactions should be excluded, and if only cross sections are
          being considered, then the irrelevant reactions should be excluded.
        """
        self._index_name = integral_index_name
        self._plot_redundant = plot_redudant
    
    def _create_plot(self, contributions: Union[Dict[str, ufloat], Dict[str, Dict[str, ufloat]]], nested: bool):
        """Creates a pie chart of the given contributions to the integral index.
        
        Parameters
        ----------
        contributions
            * If ``nested`` is ``False``, then this should be a dictionary of the form ``{nuclide: contribution}``, where 
              contribution is a ``ufloat`` object representing the contribution of the nuclide to the integral index.
            * If ``nested`` is ``True``, then this should be a dictionary of the form ``{nuclide: {reaction: contribution}}``,
              where contribution is a ``ufloat`` object representing the contribution of the nuclide to the integral index 
              through the given reaction.
        nested
            Wether the contributions are on a reaction-wise basis or not."""
        
        self._nested = nested
        self._fig, self._axs = plt.subplots(figsize=(10, 10))
        self._textprops = dict(va="center", rotation_mode = 'anchor', fontsize=12)
        if nested:
            self._nested_pie_chart(contributions)
        else:
            self._pie_chart(contributions)

        self._style()

    def _add_to_subplot(self, fig, position):
        return fig.add_subplot(position, sharex=self.ax, sharey=self.ax)

    def _get_plot(self) -> Tuple[Figure, Axes]:
        return self._fig, self._axs

    def _nested_pie_chart(self, contributions: Dict[str, Dict[str, ufloat]]):
        """Create a pie chart of the contributions to the integral index on a nuclide-reaction-wise basis.

        Parameters
        ----------
        contributions
            A dictionary of the form ``{nuclide: {reaction: contribution}}``, where contribution is a ``ufloat`` object
            representing the contribution of the nuclide to the integral index through the given reaction."""
        
        def blend_colors(color1, color2, alpha):
            return np.array( [ alpha * c1 + (1 - alpha) * c2 for c1, c2 in zip(color1, color2 ) ] )
        
        # Create a nested ring chart
        nuclide_colors = plt.get_cmap('rainbow')(np.linspace(0, 1, len(contributions.keys())))
        nuclide_colors = [ blend_colors(color, [1, 1, 1, 1], 0.8) for color in nuclide_colors ]

        nuclide_totals = { nuclide: sum(contribution.n for contribution in contributions[nuclide].values()) \
                        for nuclide in contributions }
        
        # Sort nuclides by absolute magnitude of their total contributions
        nuclide_labels, nuclide_totals_sorted = zip(*sorted(nuclide_totals.items(), key=lambda x: abs(x[1]), reverse=True))

        # Now, deal with negative values

        nuclides_with_opposite_sign_contributions = []
        for nuclide, contribution in contributions.items():
            contribution_values = [contribution[reaction].n for reaction in contribution]
            if not (all(v >= 0 for v in contribution_values) or all(v <= 0 for v in contribution_values)):
                nuclides_with_opposite_sign_contributions.append(nuclide)
            
        # For nuclides with opposite sign contributions, we distinguish the positive and negative contributions
        # by coloring some of the inner ring a lighter color to indicate the negative contributions in the outer ring
        wedge_widths = list(nuclide_totals_sorted)
        inner_wedge_hatches = [None] * len(wedge_widths)

        if len(nuclides_with_opposite_sign_contributions) > 0:
            for nuclide in nuclides_with_opposite_sign_contributions:
                # First, determine the fraction of the contributions that are opposite (in sign) to the total
                total_sign = np.sign(nuclide_totals[nuclide])
                
                # Now, we want to plot the "lost" wedge width in white, i.e. the width lost from cancellations between the
                # positive and negative contributions. This will be colored a lighter color. The absolute sum of the
                # contributions represents the wedge width if there were no cancellations, so the total wedge width
                # minus the absolute sum of the contributions is "lost" wedge width.

                absolute_sum_of_contributions = sum(np.abs(contribution.n) for contribution in contributions[nuclide].values())
                
                # NOTE the sign function is needed to handle the case when the nuclide total is negative
                lost_wedge_width = absolute_sum_of_contributions - total_sign * nuclide_totals[nuclide]

                # Now, insert the lost wedge width into the wedge widths list right after the nuclide
                nuclide_index = list(nuclide_labels).index(nuclide)
                wedge_widths.insert(nuclide_index + 1, lost_wedge_width)
                nuclide_labels = list(nuclide_labels)
                nuclide_labels.insert(nuclide_index + 1, '')
                
                # The color of the lost wedge width will be a blend of the nuclide color and white
                white_color = np.array([1, 1, 1, 1])
                opacity = 1.0
                blended_color = blend_colors(white_color, nuclide_colors[nuclide_index], opacity)
                nuclide_colors = np.insert(nuclide_colors, nuclide_index + 1, blended_color, axis=0)
                
                # Add hatches to the negative total sum wedge
                if nuclide_totals[nuclide] < 0:
                    inner_wedge_hatches[nuclide_index] = '//'

        # Now make everything positive for the pie chart
        wedge_widths = np.abs(wedge_widths)

        # Plot the inner ring for nuclide totals
        inner_ring, _ = self._axs.pie(wedge_widths, radius=0.7, labels=nuclide_labels, \
                                colors=nuclide_colors, labeldistance=0.4, textprops={'fontsize': 8}, \
                                    rotatelabels=True, wedgeprops=dict(width=0.4, edgecolor='w'))

        # Add hatches to the negative total sum wedges
        for wedge, hatch in zip(inner_ring, inner_wedge_hatches):
            if hatch:
                wedge.set_hatch(hatch)

        # Plot the outer ring for reaction-specific contributions
        outer_labels = []
        outer_colors = []
        outer_sizes = []
        outer_hatches = []
        nuclide_colors = plt.get_cmap('rainbow')(np.linspace(0, 1, len(contributions.keys())))
        nuclide_colors = [ blend_colors(color, [1, 1, 1, 1], 0.8) for color in nuclide_colors ]
        for i, nuclide in enumerate( [ nuclide_label for nuclide_label in nuclide_labels if nuclide_label != '' ] ):
            blended_color = blend_colors(nuclide_colors[i], [1, 1, 1, 1], 0.8)
            
            # Sort reactions by absolute magnitude of their contributions
            reactions = contributions[nuclide]
            sorted_reactions = sorted(reactions.items(), key=lambda x: abs(x[1].n), reverse=True)
            
            for j, (reaction, contribution) in enumerate(sorted_reactions):
                outer_labels.append(reaction)
                
                outer_colors.append(blended_color)
                outer_sizes.append(np.abs(contribution.n))
                
                if contribution.n < 0:
                    outer_hatches.append('//')
                else:
                    outer_hatches.append(None)

        outer_ring, _ = self._axs.pie(outer_sizes, radius=1, labels=outer_labels, labeldistance=0.7, colors=outer_colors, \
                textprops={'fontsize': 8}, startangle=inner_ring[0].theta1, counterclock=True, \
                    rotatelabels=True, wedgeprops=dict(width=0.4, edgecolor='w'))

        # Add hatches to the negative contribution wedges
        for wedge, hatch in zip(outer_ring, outer_hatches):
            if hatch:
                wedge.set_hatch(hatch)
        
    def _pie_chart(self, contributions: Dict[str, ufloat]):
        """Create a pie chart of the contributions to the integral index on a nuclide-wise basis.
        
        Parameters
        ----------
        contributions
            A dictionary of the form ``{nuclide: contribution}``, where contribution is a ``ufloat`` object representing the
            contribution of the nuclide to the integral index."""
        labels, values = zip(*sorted(contributions.items(), key=lambda x: abs(x[1].n), reverse=True))
        values = [values.n for values in values]
        
        # Determining hatching patterns: empty string for positive, cross-hatch for negative
        hatches = ['//' if contributions[key].n < 0 else '' for key in labels]

        # Creating the pie chart  
        wedges, _ = self._axs.pie(np.abs(values), labels=labels, startangle=90, rotatelabels =True, textprops=self._textprops)

        # Applying hatching patterns to the wedges
        for wedge, hatch in zip(wedges, hatches):
            wedge.set_hatch(hatch)

    def _style(self):
        if self._plot_redundant and self._nested:
            title_text = f'Contributions to {self._index_name} (including redundant/irrelvant reactions)'
        else:
            title_text = f'Contributions to {self._index_name}'
        self._axs.grid(True, which='both', axis='y', color='gray', linestyle='-', linewidth=0.5)
        self._axs.set_title(title_text, pad=80) # Pad set to avoid overlap with labels

class _InteractivePiePlotter(_Plotter):
    """A class for creating interactive pie charts of contributions to integral indices."""
    _index_name: str
    """The name of the integral index whose contributions are being plotted (e.g. ``'E'`` or ``'c_k'``)."""
    _plot_redundant: bool
    """Whether to plot redundant/irrelevant reactions in the pie chart."""
    def __init__(self, integral_index_name: str, plot_redundant: bool=False, **kwargs: dict):
        """Initializes a pie chart of the contributions to the given integral index.
        
        Parameters
        ----------
        integral_index_name
            The name of the integral index whose contributions are to be plotted.
        plot_redundant
            Wether to include redundant/irrelevant reactions in the plot. NOTE: this only applies to nested plots, and
            only affects the plot title; it is expected that the provided data is consistent with the flag.
        kwargs
            Additional keyword arguments to control the behavior of the interactive legend.

            - interactive_legend (bool)
                Wether to include an interactive legend in the plot. Default is ``True``.
        Notes
        -----
        * Redundant reactions are defined as those which are derived from other reactions, e.g. 'total' and 'capture' reactions
          in SCALE.
        * Irrelevant reactions are defined as those which are not directly cross sections (but rather other nuclear data 
          parameters), e.g. 'chi' in SCALE.
        * A flag for including/excluding redundant/irrelevant reactions was provided since, if the user is expecting the
          the contributions to add up nicely, then the redundant reactions should be excluded, and if only cross sections are
          being considered, then the irrelevant reactions should be excluded.
        """
        # Check if the user wants an interactive legend
        if 'interactive_legend' in kwargs.keys():
            self.interactive_legend = kwargs['interactive_legend']
        else:
            self.interactive_legend = True
        
        self._index_name = integral_index_name
        self._plot_redundant = plot_redundant

    def _create_plot(self, contributions: Union[Dict[str, ufloat], Dict[str, Dict[str, ufloat]]], nested: bool):
        """Creates an interactive pie chart of the given contributions to the integral index.
        
        Parameters
        ----------
        contributions
            * If ``nested`` is ``False``, then this should be a dictionary of the form ``{nuclide: contribution}``, where 
              contribution is a ``ufloat`` object representing the contribution of the nuclide to the integral index.
            * If ``nested`` is ``True``, then this should be a dictionary of the form ``{nuclide: {reaction: contribution}}``,
              where contribution is a ``ufloat`` object representing the contribution of the nuclide to the integral index 
              through the given reaction.
        nested
            Wether the contributions are on a reaction-wise basis or not."""
        self.fig = make_subplots()

        # Prepare data for the sunburst chart
        self.nested = nested
        if nested:
            df = self._create_nested_sunburst_data(contributions)
        else:
            df = self._create_sunburst_data(contributions)
        
        # Create a sunburst chart
        self.fig = px.sunburst(
            data_frame=df,
            names='labels',
            parents='parents',
            ids='ids',
            values='normalized_values',
            custom_data=['values', 'uncertainties']
        )

        # Update hovertemplate with correct syntax
        self.fig.update_traces(
            hovertemplate=(
                "<b>%{label}</b><br>"
                "Value: %{customdata[0]:1.4E} +/- %{customdata[1]:1.4E}"  # Corrected format specifiers
                "<extra></extra>"  # This hides the trace info
            )
        )

        # Now style the plot
        self._style()

        self.fig.update_layout(
            autosize=True,
            width=None,  # Removes fixed width
            height=None,  # Removes fixed height
            margin=dict(l=5, r=5, t=30, b=5)
        )

        if self.interactive_legend:
            self.fig = InteractivePieLegend(self.fig, df)


    
    def _add_to_subplot(self, fig, position):
        if self.interactive_legend:
            raise ValueError("Interactive legends are not supported when adding to a subplot")
        else:
            for trace in self.fig.data:
                fig.add_trace(trace, row=position[0], col=position[1])
            return fig

    def _get_plot(self) -> Union[plotly.graph_objs.Figure, InteractivePieLegend]:
        return self.fig

    def _create_sunburst_data(self, contributions: Dict[str, ufloat]) -> pd.DataFrame:
        """Create a pandas dataframe for a (not nested) sunburst chart of contributions.
        
        Parameters
        ----------
        contributions
            A dictionary of the form ``{nuclide: contribution}``, where contribution is a ``ufloat`` object representing the
            contribution of the nuclide to the integral index.
        
        Returns
        -------
            The dataframe for creating the sunburst chart."""
        data = {
            'labels': [], 
            'ids': [], 
            'parents': [], 
            'values': [], 
            'uncertainties': [],
            'normalized_values': [],
            'nuclide': []
        }

        abs_sum_of_nuclide_totals = sum( abs(contribution.n) for contribution in contributions.values())

        for nuclide, nuclide_total in contributions.items():
            # Caclulate the nuclide total, and the positive and negative contributions
            norm_nuclide_total = abs(nuclide_total) / abs_sum_of_nuclide_totals

            # Add the nuclide as a parent
            data['labels'].append(nuclide)
            data['ids'].append(nuclide)
            data['parents'].append('')
            data['values'].append(nuclide_total.n)
            data['uncertainties'].append(nuclide_total.s)
            data['normalized_values'].append(norm_nuclide_total.n)
            data['nuclide'].append(nuclide)

        return pd.DataFrame(data)

    def _create_nested_sunburst_data(self, contributions: Dict[str, Dict[str, ufloat]]) -> pd.DataFrame:
        """Create a pandas dataframe for a nested sunburst chart of contributions.
        
        Parameters
        ----------
        contributions
            A dictionary of the form ``{nuclide: {reaction: contribution}}``, where contribution is a ``ufloat`` object
            representing the contribution of the nuclide to the integral index through the given reaction.
        
        Returns
        -------
            The dataframe for creating the sunburst chart."""
        data = {
            'labels': [], 
            'ids': [], 
            'parents': [], 
            'values': [], 
            'uncertainties': [],
            'normalized_values': [],
            'nuclide': []
        }

        abs_sum_of_nuclide_totals = sum(sum(abs(contribution.n) for contribution in reactions.values()) \
                                    for reactions in contributions.values())

        for nuclide, reactions in contributions.items():
            # Caclulate the nuclide total, and the positive and negative contributions
            nuclide_total = sum(contribution for contribution in reactions.values())
            if abs_sum_of_nuclide_totals != 0:
                norm_nuclide_total = abs(nuclide_total) / abs_sum_of_nuclide_totals
            else:
                norm_nuclide_total = 0

            positive_contributions = { reaction: contribution for reaction, contribution in reactions.items() \
                                      if contribution.n >= 0 }
            negative_contributions = { reaction: contribution for reaction, contribution in reactions.items() \
                                      if contribution.n < 0 }
            positive_total = sum(contribution for contribution in positive_contributions.values())
            negative_total = sum(contribution for contribution in negative_contributions.values())

            # Add the nuclide as a parent
            data['labels'].append(nuclide)
            data['ids'].append(nuclide)
            data['parents'].append('')
            data['values'].append(nuclide_total.n)
            data['uncertainties'].append(nuclide_total.s)
            data['normalized_values'].append(norm_nuclide_total.n)
            data['nuclide'].append(nuclide)
    
            # --------------------------------------------------------
            # Add the positive and negative contributions as children
            # --------------------------------------------------------

            # Normalize the contributions by the absolute value of the nuclide total 
            absolute_sum = positive_total + abs(negative_total)
            if absolute_sum != 0:
                normalization_factor = abs(norm_nuclide_total) / absolute_sum
            else:
                normalization_factor = 0

            # Positive contributions
            if positive_total != 0:
                norm_positive_total = positive_total * normalization_factor
                data['labels'].append('Positive')
                data['ids'].append(f"{nuclide}-Positive")
                data['parents'].append(nuclide)
                data['values'].append(positive_total.n)
                data['uncertainties'].append(positive_total.s)
                data['normalized_values'].append( norm_positive_total.n )
                data['nuclide'].append(nuclide)
            else:
                norm_positive_total = 0

            # Negative contributions
            if negative_total != 0:
                norm_negative_total = abs(negative_total) * normalization_factor
                data['labels'].append('Negative')
                data['ids'].append(f"{nuclide}-Negative")
                data['parents'].append(nuclide)
                data['values'].append(negative_total.n)
                data['uncertainties'].append(negative_total.s)
                data['normalized_values'].append( norm_negative_total.n )
                data['nuclide'].append(nuclide)
            else:
                norm_negative_total = 0

            # -------------------------------
            # Add the reaction contributions
            # -------------------------------
            # NOTE: Plotly express apparently has issues dealing with small numbers, so unless the contribution is
            # multiplied by a sufficiently large scale factor, the data won't be displayed correctly
            scale_factor = 10000
            for reaction, contribution in positive_contributions.items():
                # Now normalize contributions so they sum to the "normalized_positive_total
                if positive_total != 0:
                    normalization_factor = norm_positive_total / positive_total
                else:
                    normalization_factor = 0
                norm_reaction_contribution = contribution.n * normalization_factor
                
                if contribution.n != 0:
                    data['labels'].append(reaction)
                    data['ids'].append(f"{nuclide}-{reaction}")
                    data['parents'].append(f"{nuclide}-Positive")
                    data['values'].append(contribution.n)
                    data['uncertainties'].append(contribution.s)
                    data['normalized_values'].append(scale_factor*norm_reaction_contribution.n)
                    data['nuclide'].append(nuclide)

            for reaction, contribution in negative_contributions.items():
                # Now normalize contributions so they sum to the "normalized_negative_total"
                normalization_factor = norm_negative_total / abs(negative_total)
                norm_reaction_contribution = abs(contribution.n) * normalization_factor

                if contribution.n != 0:
                    data['labels'].append(reaction)
                    data['ids'].append(f"{nuclide}-{reaction}")
                    data['parents'].append(f"{nuclide}-Negative")
                    data['values'].append(contribution.n)
                    data['uncertainties'].append(contribution.s)
                    data['normalized_values'].append(scale_factor*norm_reaction_contribution.n)
                    data['nuclide'].append(nuclide)


        return pd.DataFrame(data)

    def _style(self):
        if self._plot_redundant and self.nested:
            title_text = f'Contributions to {self._index_name} (including redundant/irrelvant reactions)'
        else:
            title_text = f'Contributions to {self._index_name}'
        self.fig.update_layout(title_text=title_text, title_x=0.5)  # 'title_x=0.5' centers the title


class InteractivePieLegend:
    """A class for creating an interactive legend for a sunburst chart."""
    fig: plotly.graph_objs.Figure
    """The sunburst chart for which the interactive legend is being created."""
    df: pd.DataFrame
    """The dataframe used to create the sunburst chart."""
    _app: Flask
    """The Flask webapp that will display the interactive legend."""
    def __init__(self, fig: plotly.graph_objs.Figure, df: pd.DataFrame):
        """Create a flask webapp that will display an interactive legend for the sunburst chart.
        
        Parameters
        ----------
        fig
            The sunburst chart for which the interactive legend is being created.
        df
            The dataframe used to create the sunburst chart."""
        self.fig = fig
        self.df = df
        self._app = Flask(__name__)

        @self._app.route('/shutdown', methods=['POST'])
        def shutdown():
            """Function to shutdown the server"""
            os.kill(os.getpid(), signal.SIGINT)  # Send the SIGINT signal to the current process
            return 'Server shutting down...'

        @self._app.route('/')
        def show_sunburst() -> str:
            """Function to display the sunburst chart with an interactive legend
            
            Returns
            -------
                The HTML content for the sunburst chart with the interactive legend."""
            # Extract root nodes (nodes without parents)
            root_nodes = self.df[self.df['parents'] == '']

            # Generate a unique ID for the container
            container_id = f"container-{uuid.uuid4()}"

            # Generate legend HTML with a title
            legend_html = f'<div id="{container_id}-legend" style="border: 2px solid black; padding: 10px;"><h3 style="margin-top: 0; text-align: center;">Legend</h3>\n'
            for _, row in root_nodes.iterrows():
                legend_html += f'<div class="{container_id}-legend-item" style="cursor: pointer; margin-bottom: 5px;" data-target="{row["ids"]}">{row["ids"]}: {row["values"]:1.4E}</div>\n'
            legend_html += '</div>\n'

            # JavaScript for interactivity and shutdown
            script_html = f"""
            <script>
            window.addEventListener('beforeunload', (event) => {{
                navigator.sendBeacon('/shutdown');
            }});
            document.addEventListener('DOMContentLoaded', function () {{
                const legendItems = document.querySelectorAll('.{container_id}-legend-item');
                legendItems.forEach(item => {{
                    item.addEventListener('mouseenter', function() {{
                        const target = this.getAttribute('data-target');
                        const paths = document.querySelectorAll('path.surface');
                        paths.forEach(path => {{
                            const labelText = path.nextElementSibling ? path.nextElementSibling.textContent : "";
                            if (labelText.includes(target)) {{
                                path.style.opacity = 0.5; // Highlight by changing opacity
                            }}
                        }});
                    }});
                    item.addEventListener('mouseleave', function() {{
                        const paths = document.querySelectorAll('path.surface');
                        paths.forEach(path => {{
                            path.style.opacity = 1; // Reset opacity
                        }});
                    }});
                    item.addEventListener('click', function() {{
                        const target = this.getAttribute('data-target');
                        const paths = document.querySelectorAll('path.surface');
                        paths.forEach(path => {{
                            const labelText = path.nextElementSibling ? path.nextElementSibling.textContent : "";
                            if (labelText.includes(target)) {{
                                path.dispatchEvent(new MouseEvent('click', {{ 'view': window, 'bubbles': true, 'cancelable': true }}));
                            }}
                        }});
                    }});
                }});
                // Force Redraw/Reflow
                setTimeout(() => {{
                    window.dispatchEvent(new Event('resize'));
                }}, 100); // Delay may be adjusted based on actual rendering time
            }});
            </script>
            """

            # Save the chart with interactivity and layout adjustments
            fig_html = self.fig.to_html(full_html=False, include_plotlyjs='cdn')
            full_html = f"""
            <!DOCTYPE html>
            <html>
            <head>
            <title>Interactive Sunburst Chart</title>
            <style>
                #{container_id} {{
                    display: flex;
                    flex-direction: row; /* Align children horizontally */
                    height: 100%;
                    width: 100%; /* Ensure the container takes full width */
                    margin: 0;
                    font-family: Arial, sans-serif;
                }}
                #{container_id} > div {{
                    display: flex;
                    justify-content: space-between; /* Space out the chart and legend */
                    align-items: flex-start; /* Align items at the start of the cross axis */
                    width: 100%;
                    overflow: hidden; /* Hide overflow to prevent breaking layout */
                }}
                #{container_id}-chart {{
                    flex: 1 1 70%; /* Allow chart to grow and shrink but base at 70% width */
                    padding: 10px;
                }}
                #{container_id}-legend {{
                    flex: 0 1 30%; /* Start with 30% width but allow shrinking */
                    padding: 5px;
                    max-height: calc(100vh - 20px); /* Limit height to viewport height minus some margin */
                    overflow: auto; /* Scroll internally if content overflows */
                }}
            </style>
            </head>
            <body>
            <div id="{container_id}">
                <div>
                    <div id="{container_id}-chart">{fig_html}</div>
                    <div id="{container_id}-legend">{legend_html}</div>
                </div>
            </div>
            {script_html}
            </body>
            </html>
            """

            return render_template_string(full_html)
        
    def _open_browser(self, port: int):
        """Open the browser to display the interactive sunburst chart

        Parameters
        ----------
        port
            The port at which the Flask server is running.
        """
        print(f"Now running at http://localhost:{port}/")
        webbrowser.open(f"http://localhost:{port}/")

    def show(self, open_browser: bool=True, silent: bool=False):
        """Start the Flask server and open the browser to display the interactive sunburst chart
        
        Parameters
        ----------
        open_browser
            Whether to open the browser automatically to display the chart.
        silent
            Whether to suppress Flask's startup and runtime messages."""
        # Suppress Flask's startup and runtime messages by redirecting them to dev null
        log = open(os.devnull, 'w')
        if silent:
            sys.stdout = log
        sys.stderr = log

        port = _find_free_port()
        if not config.generating_docs:
            if open_browser:
                threading.Timer(1, self._open_browser(port)).start()

            
            self._app.run(host='localhost', port=port)

    def serve(self):
        """Start the Flask server to display the interactive sunburst chart without a browser tab."""
        port = _find_free_port()
        log = open(os.devnull, 'w')
        # sys.stdout = log
        # sys.stderr = log

        # Run the Flask application in a separate thread
        thread = threading.Thread(target=lambda: self._app.run(host='localhost', port=port))
        print(f"Now running at http://localhost:{port}/")
        thread.daemon = True  # This ensures thread exits when main program exits
        thread.start()

    def write_html(self, filename: Optional[Union[str, Path]]=None) -> Union[None, str]:
        """Write the HTML content of the interactive sunburst chart to a file.
        
        Parameters
        ----------
        filename
            The path of the file to which the HTML content will be written. If not provided, the content will be returned.
            
        Returns
        -------
            If ``filename`` is not provided, the HTML content of the interactive sunburst chart."""
        with self._app.test_client() as client:
            response = client.get('/')
            html_content = response.data.decode('utf-8')
            if filename is None:
                return html_content
            else:
                with open(filename, 'w') as f:
                    f.write(html_content)

    def save_state(self, filename: Optional[Union[str, Path]]=None
                   ) -> Union[None, Dict[str, Union[plotly.graph_objs.Figure, pd.DataFrame]]]:
        """Save the figure (with interactive legend) as a pickle that can be deserialized for plotting later with full
        interactivity.
        
        Parameters
        ----------
        filename
            Path to the file to store the pickle of the plot object (should have a ``.pkl`` extension).

        Returns
        -------
            If ``filename`` is not provided, a dictionary containing the figure and dataframe used to create the plot. This
            data can be used to researialize the plot later.
            
        Notes
        -----
        This data can be loaded later using the :meth:`load_state` method."""
        state = {
            'fig': self.fig,
            'df': self.df,
        }
        if filename is None:
            return state
        else:
            with open(filename, 'wb') as f:
                pickle.dump(state, f)

    @classmethod
    def load_state(cls, filename: Optional[Union[str, Path]]=None, data_dict: Optional[dict]=None) -> InteractivePieLegend:
        """Researilize an :class:`InteractivePieLegend` plot from a pickle file or a data dictionary (produced by 
        :meth:`save_state`).
        
        Parameters
        ----------
        filename
            Path to the file containing the pickle of the plot object.
        data_dict
            A dictionary containing the figure and dataframe used to create the plot.
        
        Returns
        -------
            A new instance of the :class:`InteractivePieLegend` class with the same state as the original instance whose data
            was provided.
        """
        if filename is None and data_dict is None:
            raise ValueError("Either a filename or a data dictionary must be provided")
        if filename is not None:
            with open(filename, 'rb') as f:
                state = pickle.load(f)
        else:
            state = data_dict
        fig = state['fig']
        df = state['df']
        instance = cls(fig, df)
        return instance
    
    def to_image(self, filename: Union[str, Path]) -> None:
        """Write the initial state of the interactive plot to an image file. This function saves the plot as an 
        image by using selenium webdriver.
        
        Parameters
        ----------
        filename
            Name of the image file to save the matrix plot to. The file extension should be ``'.png'``, ``.jpg``, etc."""
        with tempfile.NamedTemporaryFile(delete=False) as f:
            html_filename = f.name + '.html'
        self.write_html(html_filename)
        _capture_html_as_image(html_filename, filename)
        os.remove(html_filename)


import re
import numpy as np
from uncertainties import ufloat
from pathlib import Path
from typing import Callable
import functools
from typing import List, Union
import tempfile
from string import Template
import subprocess
from tsunami_ip_utils.readers import read_integral_indices
import os

def _isotope_reaction_list_to_nested_dict(isotope_reaction_list, field_of_interest):
    """Converts a list of dictionaries containing isotope-reaction pairs (and some other key that represents a value of
    interest, e.g. an sdf profile or a contribution) to a nested dictionary
    
    Parameters
    ----------
    - isotope_reaction_list: list of dict, list of dictionaries containing isotope-reaction pairs and some other key
    - field_of_interest: str, the key in the dictionary that represents the value of interest

    Returns
    -------
    - nested_dict: dict, nested dictionary containing the isotope-reaction pairs and the value of interest"""

    isotope_reaction_dict = {}

    def get_atomic_number(isotope):
        matches = re.findall(r'\d+', isotope)
        return int(matches[0] if matches else -1) # Return -1 if no atomic number is found, these isotopes will be sorted last
                                                  # Should only be applicable to carbon in ENDF 7.1
    
    # Sort isotopes by atomic number so plots will have similar colors across different calls
    all_isotopes = list(set([isotope_reaction['isotope'] for isotope_reaction in isotope_reaction_list]))
    all_isotopes.sort(key=get_atomic_number)
    isotope_reaction_dict = { isotope: {} for isotope in all_isotopes }

    for isotope_reaction in isotope_reaction_list:
        isotope = isotope_reaction['isotope']
        reaction = isotope_reaction['reaction_type']
        value = isotope_reaction[field_of_interest]

        isotope_reaction_dict[isotope][reaction] = value

    return isotope_reaction_dict

def _filter_redundant_reactions(data_dict, redundant_reactions=['chi', 'capture', 'nubar', 'total']):
    """Filters out redundant reactions from a nested isotope-reaction dictionary
    
    Parameters
    ----------
    - data_dict: dict, nested dictionary containing isotope-reaction pairs
    - redundant_reactions: list of str, list of reactions to filter out"""
    return { isotope: { reaction: data_dict[isotope][reaction] for reaction in data_dict[isotope] \
                        if reaction not in redundant_reactions } for isotope in data_dict }

def _filter_by_nuclie_reaction_dict(data_dict, nuclide_reactions):
    """Filters out isotopes that are not in the nuclide_reactions dictionary
    
    Parameters
    ----------
    - data_dict: dict, nested dictionary containing isotope-reaction pairs
    - nuclide_reactions: dict, dictionary containing isotopes and their reactions"""
    return {nuclide: {reaction: xs for reaction, xs in reactions.items() if reaction in nuclide_reactions[nuclide]} \
                        for nuclide, reactions in data_dict.items() if nuclide in nuclide_reactions.keys()}


def _parse_ufloats(array_of_strings):
    """Parses a 2D array of strings into a 2D array of ufloats, assuming zero uncertainty if '+/-' is not found. """
    def to_ufloat(s):
        if isinstance(s, float):
            return ufloat(s, 0)
        
        parts = s.split('+/-')
        if len(parts) == 2:
            value, error = parts
        else:
            value = s
            error = 0
        return ufloat(float(value), float(error))
    
    return np.vectorize(to_ufloat)(array_of_strings)

def _convert_paths(func: Callable) -> Callable:
    """
    Decorator to ensure that any list argument passed to the decorated function,
    which contains strings, has those strings converted to pathlib.Path objects.
    """
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        new_args = []
        for arg in args:
            if isinstance(arg, list):
                # Convert all string items in the list to Path objects
                new_arg = [Path(item) if isinstance(item, str) else item for item in arg]
                new_args.append(new_arg)
            else:
                new_args.append(arg)
        
        new_kwargs = {k: [Path(item) if isinstance(item, str) else item for item in v] if isinstance(v, list) else v
                      for k, v in kwargs.items()}
        
        return func(*new_args, **new_kwargs)
    
    return wrapper

@_convert_paths
def _run_and_read_TSUNAMI_IP(application_filenames: Union[List[str], List[Path]], 
                            experiment_filenames: Union[List[str], List[Path]], coverx_library: str):
    """Runs TSUNAMI-IP and reads the output file using the :func:`tsunami_ip_utils.readers.read_integral_indices` function.
    
    Parameters
    ----------
    application_filenames
        List of paths to the application SDF files.
    experiment_filenames
        List of paths to the experiment SDF files.
    coverx
        The coverx library to use for TSUNAMI-IP.
        
    Returns
    -------
        Integral matrices for each integral index type. The shape of the matrices are ``(num_applications, num_experiments)``. 
        Keys are ``'c_k'``, ``'E_total'``, ``'E_fission'``, ``'E_capture'``, and ``'E_scatter'``."""
    
    current_dir = Path(__file__).parent
    template_filename = current_dir / 'input_files' / 'tsunami_ip_base_input.inp'
    with open(template_filename, 'r') as f:
        template = Template(f.read())

    # Convert filenames to strings
    application_filenames = [str(filename.absolute()) for filename in application_filenames]
    experiment_filenames  = [str(filename.absolute()) for filename in experiment_filenames]
    
    with tempfile.NamedTemporaryFile(mode='w', delete=False) as f:
        # Create the input file
        input_file = template.substitute(
            application_filenames='\n'.join(application_filenames),
            experiment_filenames='\n'.join(experiment_filenames),
            coverx_library=coverx_library
        )
        f.write(input_file)
        input_filename = f.name

    # Run the TSUNAMI-IP calculation
    process = subprocess.Popen( ['scalerte', input_filename], cwd=str( Path( input_filename ).parent ) )
    process.wait()

    # Read the output file
    output_filename = f"{input_filename}.out"
    tsunami_ip_output = read_integral_indices(output_filename)

    # Clean up the temporary files
    os.remove(input_filename)
    os.remove(output_filename)

    return tsunami_ip_output




from ._bar_plot import _BarPlotter
from .pie_plot import _PiePlotter, _InteractivePiePlotter
from .scatter_plot import _ScatterPlotter, _InteractiveScatterPlotter, _InteractivePerturbationScatterPlotter
from .matrix_plot import _interactive_matrix_plot
from .plot_utils import _determine_plot_type
from tsunami_ip_utils.integral_indices import _add_missing_reactions_and_nuclides
import numpy as np
from uncertainties import ufloat, unumpy
from typing import List, Dict, Tuple, Union, Optional
from pathlib import Path
import pandas as pd
import matplotlib.pyplot as plt
from tsunami_ip_utils._utils import _parse_ufloats
from tsunami_ip_utils import config
from matplotlib.figure import Figure
from matplotlib.axes import Axes
from tsunami_ip_utils.viz.pie_plot import InteractivePieLegend
import plotly
from tsunami_ip_utils.viz.scatter_plot import InteractiveScatterLegend
from tsunami_ip_utils.viz.scatter_plot import EnhancedPlotlyFigure
from tsunami_ip_utils.viz.matrix_plot import InteractiveMatrixPlot

def contribution_plot(contributions: List[Dict], plot_type: str='bar', integral_index_name: str='E', 
                      plot_redundant_reactions: bool=True, **kwargs: dict
                      ) -> Union[Tuple[Figure, Axes], plotly.graph_objs.Figure, InteractivePieLegend]:
    """Plots the contributions to an arbitrary similarity parameter for a single experiment application pair
    
    Parameters
    ----------
    contributions
        List of dictionaries containing the contributions to the similarity parameter for each
        nuclide or nuclide-reaction pair.
    plot_type
        Type of plot to create. Default is ``'bar'`` which creates a matplotlib bar plot. Other options are:
        
        - ``'pie'``
            Creates a ``matplotlib`` pie chart.
        - ``'interactive_pie'``
            Creates an interactive Plotly sunburst chart.
    integral_index_name
        Name of the integral index being plotted. Default is ``'E'``.
    plot_redundant_reactions
        Whether to plot redundant reactions (or irrelevant reactions) when considering
        nuclide-reaction-wise contributions. Default is ``True``.
    kwargs
        Additional keyword arguments to pass to the plotting function.
        
    Returns
    -------
        * If ``plot_type`` is ``'bar'``, returns a tuple containing the matplotlib figure and axes objects. These
          can be used to further style the plot or save it to a file. Can be plotted by calling ``plt.show()``.
        * If ``plot_type`` is ``'pie'``, returns a Plotly figure object which can be displayed by calling ``fig.show()``
        * If ``plot_type`` is ``'interactive_pie'``, returns a :class:`.InteractivePieLegend` object which can be displayed by
          calling ``fig.show()``"""

    # Determine if the contributions are nuclide-wise or nuclide-reaction-wise
    contributions, nested_plot = _determine_plot_type(contributions, plot_redundant_reactions)

    plotters = {
        'bar': _BarPlotter(integral_index_name, plot_redundant_reactions, **kwargs),
        'pie': _PiePlotter(integral_index_name, plot_redundant_reactions, **kwargs),
        'interactive_pie': _InteractivePiePlotter(integral_index_name, plot_redundant_reactions, **kwargs)
    }
    
    # Get the requested plotter
    plotter = plotters.get(plot_type)
    if plotter is None:
        raise ValueError("Unsupported plot type")

    # Create the plot and style it
    plotter._create_plot(contributions, nested_plot)

    return plotter._get_plot()


def correlation_plot(application_contributions: List[dict], experiment_contributions: List[dict], plot_type: str='scatter', 
                     integral_index_name: str='E', plot_redundant_reactions: bool=True, **kwargs: dict
                     ) -> Union[Tuple[Figure, Axes], EnhancedPlotlyFigure, InteractiveScatterLegend]:
    """Creates a correlation plot for a given application-experiment pair for which the contributions to the similarity
    parameter are given.
    
    Parameters
    ----------
    application_contributions
        List of dictionaries containing the contributions to the similarity parameter
        for the application.
    experiment_contributions
        List of dictionaries containing the contributions to the similarity parameter
        for the experiment.
    plot_type
        Type of plot to create. Default is ``'scatter'`` which creates a matplotlib scatter plot. Other options
        are ``'interactive_scatter'``, which creates a Plotly scatter plot.
    integral_index_name
        Name of the integral index being plotted. Default is ``'E'``
    plot_redundant_reactions
        Whether to plot redundant reactions (or irrelevant reactions) when considering
        nuclide-reaction-wise contributions. Default is ``True``.
    kwargs
        Additional keyword arguments to pass to the plotting function.

        - ``interactive_legend``: bool
            Whether to include an interactive legend in the plot. Default is ``False``. Only applicable if
            ``plot_type`` is ``'interactive_scatter'``.
        
    Returns
    -------
        * If ``plot_type`` is ``'scatter'``, returns a tuple containing the matplotlib figure and axes objects. These
          can be used to further style the plot or save it to a file. Can be plotted by calling ``plt.show()``.
        * If ``plot_type`` is ``'interactive_scatter'`` and ``interactive_legend`` is ``False``, returns a Plotly figure 
          object which can be displayed by calling ``fig.show()``
        * If ``plot_type`` is ``'interactive_scatter'`` and ``interactive_legend`` is ``True``, returns a 
          :class:`.InteractiveScatterLegend` object which can be displayed by calling ``fig.show()``
          
    Notes
    -----
    Technically, if ``plot_type`` is ``interactive_scatter`` and ``interactive_legend`` is ``False``, the function will
    return a :class:`.EnhancedPlotlyFigure` object, which is a wrapper of Plotly's ``Figure`` class, which has additional
    attributes related to the statistics of the scatter plot."""

    # Determine if the contributions are nuclide-wise or nuclide-reaction-wise
    application_contributions, application_nested = _determine_plot_type(application_contributions, plot_redundant_reactions)
    experiment_contributions, experiment_nested = _determine_plot_type(experiment_contributions, plot_redundant_reactions)

    if application_nested != experiment_nested:
        raise ValueError("Application and experiment contributions must have the same nested structure")
    else:
        nested = application_nested # They are be the same, so arbitrarily choose one

    # Get the list of isotopes for which contributions are available
    isotopes = list(set(application_contributions.keys()).union(experiment_contributions.keys()))

    all_reactions = _add_missing_reactions_and_nuclides(application_contributions, experiment_contributions, isotopes, mode='contribution')

    # Now convert the contributions for the application and experiment into a list of x, y pairs for plotting
    contribution_pairs = []
    if nested:
        for isotope in isotopes:
            for reaction in all_reactions:
                contribution_pairs.append((application_contributions[isotope][reaction], \
                                           experiment_contributions[isotope][reaction]))
    else:
        for isotope in isotopes:
            contribution_pairs.append((application_contributions[isotope], experiment_contributions[isotope]))

    plotters = {
        'scatter': _ScatterPlotter(integral_index_name, plot_redundant_reactions, nested, **kwargs),
        'interactive_scatter': _InteractiveScatterPlotter(integral_index_name, plot_redundant_reactions, nested, **kwargs)
    }
    
    # Get the requested plotter
    plotter = plotters.get(plot_type)
    if plotter is None:
        raise ValueError(f"Unsupported plot type: {plot_type}")

    # Create the plot and style it
    plotter._create_plot(contribution_pairs, isotopes, all_reactions)

    return plotter._get_plot()


def perturbation_plot(points: List[Tuple[ufloat, ufloat]]) -> EnhancedPlotlyFigure:
    """Plots the perturbation points for a given application-experiment pair for which the perturbation points have already
    been calculated.
    
    Parameters
    ----------
    points: 
        List of tuples containing the perturbation points for the application-experiment pair.
        
    Returns
    -------
        A (enhanced) Plotly figure object which can be displayed by calling ``fig.show()``."""
    
    # Extracting data
    plotter = _InteractivePerturbationScatterPlotter()
    plotter._create_plot(points)

    return plotter._get_plot()


def matrix_plot(plot_objects_array: np.ndarray, plot_type: str, labels: Optional[Dict[str, List]]=None
                ) -> InteractiveMatrixPlot:
    """Creates a Dash app to display a matrix of plots from a numpy object array of figure objects.
    
    Parameters
    ----------
    plot_objects_array
        A 2D numpy array of figure objects.
    plot_type
        Type of plot to create. Default is ``'interactive'`` which creates an interactive matrix plot (served by a Dash app). 
        Other options are ``'static'``, which creates a static matplotlib matrix plot.
    labels
        Dictionary of lists containing the labels for the rows and columns of the matrix plot. Keys are ``'expeirments'`` and 
        ``'applications'``. Default is ``None``. Most commonly, this is the sdf filename for the given application/experiment.
        
    Returns
    -------
        * If ``plot_type`` is ``'interactive'``, returns a :class:`.InteractiveMatrixPlot` object which can be run by 
          calling ``app.show()``.
        * If ``plot_type`` is ``'static'``, raises a ``NotImplementedError``."""
    if plot_type == 'interactive':
        return _interactive_matrix_plot(plot_objects_array, labels)
    elif plot_type == 'static':
        raise NotImplementedError("Static matrix plots are not yet supported")


class BlockingFigureWrapper:
    """This class wraps a ``matplotlib.figure.Figure`` instance and blocks the execution of the program until the figure
    is closed. This is useful for displaying figures in a Jupyter notebook or other interactive environments where the
    program would otherwise continue executing and close the figure immediately, i.e. ``figure.show()`` is non-blocking. 
    This is nice for creating multiple plots programmatically via a plotting function, then returning the relevant figure 
    and axs objects, after which the user can make changes and plot or save the figures individually. This functionality
    is not allowed with ``plt.show()``, because it shows all plots created in a given session (and so will show multiple plots
    instead of just a selected one)."""
    def __init__(self, figure: Figure):
        if not isinstance(figure, Figure):
            raise ValueError("Expected a matplotlib.figure.Figure instance")
        self._figure = figure

    def show(self):
        """Display the figure and block the program until the user presses Enter."""
        # Show the figure
        self._figure.show()

        input("Press Enter to continue...")

    def __getattr__(self, name):
        """Delegate all other attribute access to the wrapped matplotlib figure instance."""
        return getattr(self._figure, name)


def generate_heatmap_from_comparison(comparison_excel_path: Union[str, Path], base_fontsize: int=6
                                     ) -> Dict[str, Tuple[plt.Figure, plt.Axes]]:
    """Generates a heatmap from a comparison excel file generated by :func:`tsunami_ip_utils.comparisons.correlation_comparison`.
    
    Parameters
    ----------
    comparison_excel_path
        Path to the comparison excel file.
    base_fontsize
        Base font size for the heatmap (there is some heuristic for scaling fontsize with matrix size). Default is 6.
    
    Returns
    -------
        A dictionary containing the matplotlib figure and axes objects for each unique header in the comparison excel file."""
    
    comparison = pd.read_excel(comparison_excel_path, header=[0,1], index_col=0)
    
    # Get unique secondary headers from level 1 of the MultiIndex
    unique_headers = comparison.columns.get_level_values(1).unique()

    plot_dict = {}
    heatmap_labels = config.COMPARISON_HEATMAP_LABELS

    # Extract a separate 2D numpy array for each unique header in level 1
    for index, header in enumerate(unique_headers):
        # Filter the DataFrame for the current header
        filtered_df = comparison.xs(header, level=1, axis=1)
        # Convert the filtered DataFrame to a numpy array
        array = _parse_ufloats(filtered_df.to_numpy())
        
        # Now plot the array
        data = unumpy.nominal_values(array)
        plt.figure(index)
        cax = plt.matshow(data, cmap='viridis')
        cbar = plt.colorbar(cax)
        cbar.set_label(heatmap_labels[header])

        # Annotate cells with values
        font_size = min(base_fontsize, 200 / max(data.shape))
        for (i, j), val in np.ndenumerate(data):
            plt.text(j, i, f"{val:.2f}", ha='center', va='center', color='white', fontsize=font_size)

        # Set axis labels and ticks
        plt.xlabel('Application')
        plt.ylabel('Experiment')
        plt.xticks(range(data.shape[1]))
        plt.yticks(range(data.shape[0]))
        plt.tick_params(axis="x", bottom=True, labelbottom=True, top=False, labeltop=False)

        plot_dict[header] = (BlockingFigureWrapper(plt.gcf()), plt.gca())

    return plot_dict


# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['tsunami_ip_utils', 'tsunami_ip_utils.viz']

package_data = \
{'': ['*'],
 'tsunami_ip_utils': ['input_files/*'],
 'tsunami_ip_utils.viz': ['css/*']}

install_requires = \
['IPython',
 'dash',
 'flask',
 'h5py',
 'kaleido',
 'matplotlib',
 'openpyxl',
 'pandas',
 'plotly>=5.22.0,<5.23.0',
 'pyprojroot',
 'pytest',
 'pytest-mpl',
 'pyyaml',
 'scipy',
 'selenium',
 'tqdm',
 'uncertainties',
 'webdriver-manager']

setup_kwargs = {
    'name': 'tsunami-ip-utils',
    'version': '0.1.0',
    'description': 'A tool for visualizing similarity data from the SCALE code: TSUNAMI-IP',
    'long_description': 'Nothing here yet',
    'author': 'Matthew Louis',
    'author_email': 'matthewlouis31@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
